"""
This file implements a pure JAX implementation for Gated Delta Networks (GDNs)

Key Implementation Considerations:
- Dynamic Shapes & Continuous Batching: vLLM mixes ragged prefill and decode tokens
  dynamically. To prevent XLA from recompiling on every varying batch size, we use
  `jax.lax.scan` to iterate over flat token streams using conditional masking, allow us
  to have a unified and statically-shaped execution path.
"""

# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
import torch
from jax.sharding import PartitionSpec as P
from torchax.interop import jax_view, torch_view
from vllm.forward_context import get_forward_context

from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.models.vllm.vllm_model_wrapper_context import \
    get_vllm_model_wrapper_context


def _l2_normalize(x: jnp.ndarray, eps: float = 1e-6) -> jnp.ndarray:
    """L2 normalize along last dimension.

    Args:
        x: input to normalize
        eps: epsilon for numerical stability

    Returns:
        normalized x
    """
    norm = jnp.sqrt(jnp.sum(x * x, axis=-1, keepdims=True) + eps)
    return x / norm


def _causal_conv1d_step(
    x_new: jnp.ndarray,
    conv_state: jnp.ndarray,
    conv_weight: jnp.ndarray,
    conv_bias: Optional[jnp.ndarray] = None,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Single-step causal conv1d for decode mode.

    During decode, instead of padding and convoluting over the whole sequence, we
    maintain a rolling KV-cache equivalent (`conv_state`) of the last `kernel_size - 1` tokens
    and compute a single step of the depthwise convolution

    Args:
        x_new: (B, 1, C) input
        conv_state: (B, T, C) state
        conv_weight: (C, 1, kernel_size) depthwise kernel (PyTorch Conv1d format)
        conv_bias: optional (C,) bias

    Returns:
        (B, 1, C) output
        (B, T, C) new state
    """
    new_state = jnp.concatenate([conv_state[:, 1:, :], x_new], axis=1)
    window = jnp.concatenate([conv_state, x_new], axis=1)

    if conv_weight.ndim == 2:
        w = conv_weight
    else:
        w = conv_weight[:, 0, :].T  # (C,1,K) → (K,C)

    out = jnp.sum(window * w[jnp.newaxis, :, :], axis=1, keepdims=True)

    if conv_bias is not None:
        out = out + conv_bias[jnp.newaxis, jnp.newaxis, :]

    return out, new_state


def _chunk_gated_delta_rule(
    query: jnp.ndarray,
    key: jnp.ndarray,
    value: jnp.ndarray,
    g: jnp.ndarray,
    beta: jnp.ndarray,
    chunk_size: int = 64,
    initial_state: Optional[jnp.ndarray] = None,
    output_final_state: bool = True,
) -> Tuple[jnp.ndarray, Optional[jnp.ndarray]]:
    """Chunked gated delta rule matching HF's torch_chunk_gated_delta_rule.

    By chunking here, we can effectively transform the purely sequential
    RNN recurrence into a block-parallel operation. It processes tokens in chunks
    and then only passes the recurrent state sequentially between chunks.

    One detail worth pointing out is that the continuous decay mask (`g`) is cumulative, so
    Applying the triangular mask *before* exponentiation is key here to prevent NaNs
    when dealing with large sequence lengths.

    Args:
        query: (B, H, T, d_k) — already L2-normed
        key: (B, H, T, d_k) — already L2-normed
        value: (B, H, T, d_v)
        g: (B, H, T) — continuous decay (negative values)
        beta: (B, H, T) — input gate (after sigmoid)
        chunk_size: chunk processing size
        initial_state: (B, H, d_k, d_v) or None

    Returns:
        output: (B, H, T, d_v)
        final_state: (B, H, d_k, d_v) or None
    """
    B, H, T, d_k = query.shape
    d_v = value.shape[-1]

    # Pad to multiple of chunk_size
    pad_size = (chunk_size - T % chunk_size) % chunk_size
    if pad_size > 0:
        query = jnp.pad(query, ((0, 0), (0, 0), (0, pad_size), (0, 0)))
        key = jnp.pad(key, ((0, 0), (0, 0), (0, pad_size), (0, 0)))
        value = jnp.pad(value, ((0, 0), (0, 0), (0, pad_size), (0, 0)))
        beta = jnp.pad(beta, ((0, 0), (0, 0), (0, pad_size)))
        g = jnp.pad(g, ((0, 0), (0, 0), (0, pad_size)))

    total_len = T + pad_size
    scale = d_k**-0.5
    query = query * scale

    v_beta = value * beta[..., jnp.newaxis]
    k_beta = key * beta[..., jnp.newaxis]

    # Reshape: (B, H, n_chunks, chunk_size, dim)
    nc = total_len // chunk_size
    query = query.reshape(B, H, nc, chunk_size, d_k)
    key = key.reshape(B, H, nc, chunk_size, d_k)
    value = value.reshape(B, H, nc, chunk_size, d_v)
    k_beta = k_beta.reshape(B, H, nc, chunk_size, d_k)
    v_beta = v_beta.reshape(B, H, nc, chunk_size, d_v)
    g = g.reshape(B, H, nc, chunk_size)

    # Cumulative g within each chunk
    g = jnp.cumsum(g, axis=-1)

    # Decay mask: tril BEFORE exp for numerical stability (matching HF)
    g_diff = g[..., :, jnp.newaxis] - g[..., jnp.newaxis, :]
    decay_mask = jnp.exp(jnp.tril(g_diff)) * jnp.tril(
        jnp.ones((chunk_size, chunk_size)))

    # Intra-chunk correction matrix
    mask = jnp.triu(jnp.ones((chunk_size, chunk_size), dtype=jnp.bool_))
    attn = -(k_beta @ jnp.swapaxes(key, -2, -1)) * decay_mask
    attn = jnp.where(mask, 0.0, attn)

    # Iterative correction (matching HF for loop exactly)
    for i in range(1, chunk_size):
        row = attn[..., i, :i]  # (B, H, nc, i)
        sub = attn[..., :i, :i]  # (B, H, nc, i, i)
        correction = jnp.sum(row[..., jnp.newaxis] * sub,
                             axis=-2)  # (B, H, nc, i)
        attn = attn.at[..., i, :i].set(row + correction)

    attn = attn + jnp.eye(chunk_size)
    value = attn @ v_beta
    k_cumdecay = attn @ (k_beta * jnp.exp(g)[..., jnp.newaxis])

    # Initialize state
    state = (jnp.zeros((B, H, d_k, d_v), dtype=jnp.float32)
             if initial_state is None else initial_state.astype(jnp.float32))

    upper2 = jnp.triu(jnp.ones((chunk_size, chunk_size)), k=1)

    all_outputs = []
    for ci in range(nc):
        q_i = query[:, :, ci]
        k_i = key[:, :, ci]
        v_i = value[:, :, ci]
        g_i = g[:, :, ci]

        intra = (q_i @ jnp.swapaxes(k_i, -2, -1)) * decay_mask[:, :, ci]
        intra = jnp.where(upper2, 0.0, intra)

        v_prime = k_cumdecay[:, :, ci] @ state
        v_new = v_i - v_prime

        attn_inter = (q_i * jnp.exp(g_i)[..., jnp.newaxis]) @ state
        chunk_out = attn_inter + intra @ v_new
        all_outputs.append(chunk_out)

        g_last = g_i[:, :, -1:]
        state = (state * jnp.exp(g_last)[..., jnp.newaxis] + jnp.swapaxes(
            k_i * jnp.exp(g_last[..., jnp.newaxis] - g_i[..., jnp.newaxis]),
            -2,
            -1,
        ) @ v_new)

    output = jnp.concatenate(all_outputs, axis=2)[:, :, :T]
    return output, state if output_final_state else None


def _recurrent_gated_delta_rule_step(
    query: jnp.ndarray,
    key: jnp.ndarray,
    value: jnp.ndarray,
    g: jnp.ndarray,
    beta: jnp.ndarray,
    state: Optional[jnp.ndarray] = None,
) -> Tuple[jnp.ndarray, Optional[jnp.ndarray]]:
    """Single-step for decode.

    Args:
        query: (B, H, T, d_k)
        key: (B, H, T, d_k)
        value: (B, H, T, d_v)
        g: (B, H, T)
        beta: (B, H, T)
        state: (B, H, d_k, d_v)

    Returns:
        output: (B, H, T, d_v)
        new_state: (B, H, d_k, d_v)
    """
    # Just use chunk_gated_delta_rule with T=1
    output, new_state = _chunk_gated_delta_rule(
        query,
        key,
        value,
        g,
        beta,
        chunk_size=1,
        initial_state=state,
        output_final_state=True,
    )
    return output, new_state


def run_jax_gdn_attention_local(
    query: jnp.ndarray,
    key: jnp.ndarray,
    value: jnp.ndarray,
    b: jnp.ndarray,
    a: jnp.ndarray,
    conv_state_q: jnp.ndarray,
    conv_state_k: jnp.ndarray,
    conv_state_v: jnp.ndarray,
    recurrent_state: jnp.ndarray,
    conv_weight_q: jnp.ndarray,
    conv_weight_k: jnp.ndarray,
    conv_weight_v: jnp.ndarray,
    conv_bias_q: jnp.ndarray,
    conv_bias_k: jnp.ndarray,
    conv_bias_v: jnp.ndarray,
    A_log: jnp.ndarray,
    dt_bias: jnp.ndarray,
    q_loc: jnp.ndarray,
    state_indices: jnp.ndarray,
    n_kq: int,
    n_v: int,
    d_k: int,
    d_v: int,
    kernel_size: int,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray],
           jnp.ndarray]:
    """Runs the local JAX GDN attention mechanism.

    Args:
        query: Query tensor of shape `(num_tokens, n_kq * d_k)`.
        key: Key tensor of shape `(num_tokens, n_kq * d_k)`.
        value: Value tensor of shape `(num_tokens, n_v * d_v)`.
        b: B tensor of shape `(num_tokens, n_v)`.
        a: A tensor of shape `(num_tokens, n_v)`.
        conv_state_q: Convolutional state for query of shape `(max_reqs, kernel_size - 1, n_kq * d_k)`.
        conv_state_k: Convolutional state for key of shape `(max_reqs, kernel_size - 1, n_kq * d_k)`.
        conv_state_v: Convolutional state for value of shape `(max_reqs, kernel_size - 1, n_v * d_v)`.
        recurrent_state: Recurrent state of shape `(max_reqs, n_v, d_k, d_v)`.
        conv_weight_q: Convolutional weight for query of shape `(n_kq * d_k, 1, kernel_size)`.
        conv_weight_k: Convolutional weight for key of shape `(n_kq * d_k, 1, kernel_size)`.
        conv_weight_v: Convolutional weight for value of shape `(n_v * d_v, 1, kernel_size)`.
        conv_bias_q: Convolutional bias for query of shape `(n_kq * d_k,)`.
        conv_bias_k: Convolutional bias for key of shape `(n_kq * d_k,)`.
        conv_bias_v: Convolutional bias for value of shape `(n_v * d_v,)`.
        A_log: Log of A parameter of shape `(n_v,)`.
        dt_bias: Delta T bias of shape `(n_v,)`.
        q_loc: Tensor of shape `(num_seqs,)` with start locations of each sequence.
        state_indices: Tensor of shape `(max_reqs,)` mapping request index to state index.
        n_kq: Number of key/query heads.
        n_v: Number of value heads.
        d_k: Dimension of key.
        d_v: Dimension of value.
        kernel_size: Convolution kernel size.

    Returns:
        A tuple containing the new states and the output.
        - A tuple of (new_conv_state_q, new_conv_state_k, new_conv_state_v, new_recurrent_state).
        - The output tensor of shape `(num_tokens, n_v * d_v)`.
    """
    # Ensure q_loc is monotonically increasing to handle padded slots
    q_loc = jnp.maximum.accumulate(q_loc)

    num_tokens = query.shape[0]
    token_idx = jnp.arange(num_tokens)
    max_reqs = state_indices.shape[0]

    req_indices = jnp.sum(token_idx[:, None] >= q_loc[None, :], axis=1) - 1
    req_indices = jnp.clip(req_indices, 0, max_reqs - 1)

    # Exclude trailing padding indices from mutating cache
    valid_mask = token_idx < q_loc[-1]

    def scan_fn(carry, xs):
        conv_state_q, conv_state_k, conv_state_v, recurrent_state_all = carry
        (
            curr_q,
            curr_k,
            curr_v,
            curr_b,
            curr_a,
            request_index,
            is_valid_token,
        ) = xs

        curr_q = curr_q[None, None, :]
        curr_k = curr_k[None, None, :]
        curr_v = curr_v[None, None, :]
        curr_b = curr_b[None, None, :]
        curr_a = curr_a[None, None, :]

        state_index = state_indices[request_index]
        cs_q = conv_state_q[state_index][None, ...]
        cs_k = conv_state_k[state_index][None, ...]
        cs_v = conv_state_v[state_index][None, ...]
        recurrent_state = recurrent_state_all[state_index][None, ...]

        # 1. Causal Conv1D
        query_conv, new_conv_state_q = _causal_conv1d_step(
            curr_q, cs_q, conv_weight_q, conv_bias_q)
        key_conv, new_conv_state_k = _causal_conv1d_step(
            curr_k, cs_k, conv_weight_k, conv_bias_k)
        value_conv, new_conv_state_v = _causal_conv1d_step(
            curr_v, cs_v, conv_weight_v, conv_bias_v)

        mixed_qkv = jnp.concatenate([query_conv, key_conv, value_conv],
                                    axis=-1)
        mixed_qkv = jax.nn.silu(mixed_qkv)

        # 2. Split Q, K, V
        B, T, _ = mixed_qkv.shape
        key_dim = n_kq * d_k
        query = mixed_qkv[..., :key_dim].reshape(B, T, n_kq, d_k)
        key = mixed_qkv[..., key_dim:key_dim * 2].reshape(B, T, n_kq, d_k)
        value = mixed_qkv[..., key_dim * 2:].reshape(B, T, n_v, d_v)

        # 3. Compute continuous decay (g) and input gate (beta)
        beta = jax.nn.sigmoid(curr_b)
        g = -jnp.exp(A_log.astype(jnp.float32)) * jax.nn.softplus(
            curr_a.astype(jnp.float32) + dt_bias.astype(jnp.float32))

        # 4. Head Expansion (GQA -> MHA logic)
        repeat_factor = n_v // n_kq
        if repeat_factor > 1:
            query = jnp.repeat(query, repeat_factor, axis=2)
            key = jnp.repeat(key, repeat_factor, axis=2)

        # Transpose to (B, H, T, dim)
        query = jnp.transpose(query, (0, 2, 1, 3)).astype(jnp.float32)
        key = jnp.transpose(key, (0, 2, 1, 3)).astype(jnp.float32)
        value = jnp.transpose(value, (0, 2, 1, 3)).astype(jnp.float32)
        beta = jnp.transpose(beta, (0, 2, 1)).astype(jnp.float32)
        g = jnp.transpose(g, (0, 2, 1)).astype(jnp.float32)

        # L2 normalize Q, K
        query = _l2_normalize(query)
        key = _l2_normalize(key)

        # 5. Delta Rule Recurrence
        output, new_recurrent_state = _recurrent_gated_delta_rule_step(
            query, key, value, g, beta, state=recurrent_state)

        # Output back to (B, T, H, d_v) -> (B, T, H * d_v)
        output = jnp.transpose(output, (0, 2, 1, 3)).astype(mixed_qkv.dtype)
        output = output.reshape(B, T, -1)

        conv_state_q = jnp.where(
            is_valid_token,
            conv_state_q.at[state_index].set(new_conv_state_q[0]),
            conv_state_q)
        conv_state_k = jnp.where(
            is_valid_token,
            conv_state_k.at[state_index].set(new_conv_state_k[0]),
            conv_state_k)
        conv_state_v = jnp.where(
            is_valid_token,
            conv_state_v.at[state_index].set(new_conv_state_v[0]),
            conv_state_v)
        recurrent_state_all = jnp.where(
            is_valid_token,
            recurrent_state_all.at[state_index].set(new_recurrent_state[0]),
            recurrent_state_all)

        return (conv_state_q, conv_state_k, conv_state_v,
                recurrent_state_all), output[0, 0]

    carry_init = (conv_state_q, conv_state_k, conv_state_v, recurrent_state)
    xs = (query, key, value, b, a, req_indices, valid_mask)

    (
        (new_conv_state_q, new_conv_state_k, new_conv_state_v,
         new_recurrent_state),
        output,
    ) = jax.lax.scan(scan_fn, carry_init, xs)

    return (new_conv_state_q, new_conv_state_k, new_conv_state_v,
            new_recurrent_state), output


def run_jax_gdn_attention(
    j_mixed_qkv: jnp.ndarray,
    j_b: jnp.ndarray,
    j_a: jnp.ndarray,
    conv_state: jnp.ndarray,
    recurrent_state: jnp.ndarray,
    j_conv_weight: jnp.ndarray,
    j_conv_bias: Optional[jnp.ndarray],
    j_A_log: jnp.ndarray,
    j_dt_bias: jnp.ndarray,
    state_indices: jnp.ndarray,
    q_loc: jnp.ndarray,
    n_kq: int,
    n_v: int,
    d_k: int,
    d_v: int,
    kernel_size: int,
    mesh: jax.sharding.Mesh,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
    """Runs the Jax GDN attention mechanism.

    Args:
        j_mixed_qkv: Input tensor of shape `(num_tokens, dim)`.
        j_b: Input tensor of shape `(num_tokens, n_v)`.
        j_a: Input tensor of shape `(num_tokens, n_v)`.
        conv_state: Convolutional state tensor of shape
          `(max_reqs, kernel_size - 1, dim)`.
        recurrent_state: Recurrent state tensor of shape
          `(max_reqs, n_v, d_k, d_v)`.
        j_conv_weight: Convolutional weight tensor of shape
          `(dim, 1, kernel_size)`.
        j_conv_bias: Optional convolutional bias tensor of shape `(dim,)`.
        j_A_log: Log of A parameter tensor of shape `(n_v,)`.
        j_dt_bias: Delta T bias tensor of shape `(n_v,)`.
        state_indices: Tensor of shape `(max_reqs,)` mapping request index to
          state index.
        q_loc: Tensor of shape `(num_seqs,)` with start locations of each
          sequence.
        n_kq: Number of key/query heads.
        n_v: Number of value heads.
        d_k: Dimension of key.
        d_v: Dimension of value.
        kernel_size: Convolution kernel size.
        mesh: The device mesh for distributed computation.

    Returns:
        A tuple containing the new states and the output.
        - A tuple of (new_conv_state, new_recurrent_state).
          - new_conv_state: `(max_reqs, kernel_size - 1, dim)`
          - new_recurrent_state: `(max_reqs, n_v, d_k, d_v)`
        - The output tensor of shape `(num_tokens, n_v * d_v)`.
    """
    key_dim = n_kq * d_k

    # NOTE: we pre-split Q, K, and V BEFORE passing to shard map,
    # otherwise, jax.shard_map splits across C dimension
    j_query = j_mixed_qkv[..., :key_dim]
    j_key = j_mixed_qkv[..., key_dim:key_dim * 2]
    j_value = j_mixed_qkv[..., key_dim * 2:]

    conv_state_query = conv_state[..., :key_dim]
    conv_state_key = conv_state[..., key_dim:key_dim * 2]
    conv_state_value = conv_state[..., key_dim * 2:]

    conv_weight_query = j_conv_weight[:key_dim, ...]
    conv_weight_key = j_conv_weight[key_dim:key_dim * 2, ...]
    conv_weight_value = j_conv_weight[key_dim * 2:, ...]

    if j_conv_bias is not None:
        conv_bias_query = j_conv_bias[:key_dim]
        conv_bias_key = j_conv_bias[key_dim:key_dim * 2]
        conv_bias_value = j_conv_bias[key_dim * 2:]
        bias_spec = P(ShardingAxisName.ATTN_HEAD)
    else:
        conv_bias_query = conv_bias_key = conv_bias_value = None
        bias_spec = None

    in_specs = (
        P(None, ShardingAxisName.ATTN_HEAD),  # j_query
        P(None, ShardingAxisName.ATTN_HEAD),  # j_key
        P(None, ShardingAxisName.ATTN_HEAD),  # j_value
        P(None, ShardingAxisName.ATTN_HEAD),  # j_b
        P(None, ShardingAxisName.ATTN_HEAD),  # j_a
        P(None, None, ShardingAxisName.ATTN_HEAD),  # conv_state_query
        P(None, None, ShardingAxisName.ATTN_HEAD),  # conv_state_key
        P(None, None, ShardingAxisName.ATTN_HEAD),  # conv_state_value
        P(None, ShardingAxisName.ATTN_HEAD, None, None),  # recurrent_state
        P(ShardingAxisName.ATTN_HEAD, None, None),  # conv_weight_query
        P(ShardingAxisName.ATTN_HEAD, None, None),  # conv_weight_key
        P(ShardingAxisName.ATTN_HEAD, None, None),  # conv_weight_value
        bias_spec,
        bias_spec,
        bias_spec,  # conv_bias_query, conv_bias_key, conv_bias_value
        P(ShardingAxisName.ATTN_HEAD),  # j_A_log
        P(ShardingAxisName.ATTN_HEAD),  # j_dt_bias
        P(),  # q_loc
        P(),  # state_indices
    )

    out_specs = (
        (
            P(None, None, ShardingAxisName.ATTN_HEAD),  # new_conv_state_q
            P(None, None, ShardingAxisName.ATTN_HEAD),  # new_conv_state_k
            P(None, None, ShardingAxisName.ATTN_HEAD),  # new_conv_state_v
            P(None, ShardingAxisName.ATTN_HEAD, None,
              None),  # new_recurrent_state
        ),
        P(None, ShardingAxisName.ATTN_HEAD),  # output
    )

    tp_size = mesh.shape[ShardingAxisName.ATTN_HEAD]

    p_run_jax_gdn_attention_local = functools.partial(
        run_jax_gdn_attention_local,
        n_kq=n_kq // tp_size,
        n_v=n_v // tp_size,
        d_k=d_k,
        d_v=d_v,
        kernel_size=kernel_size)

    mapped_fn = jax.shard_map(
        p_run_jax_gdn_attention_local,
        mesh=mesh,
        in_specs=in_specs,
        out_specs=out_specs,
    )

    (
        (new_conv_state_q, new_conv_state_k, new_conv_state_v,
         new_recurrent_state),
        output,
    ) = mapped_fn(j_query, j_key, j_value, j_b, j_a, conv_state_query,
                  conv_state_key, conv_state_value, recurrent_state,
                  conv_weight_query, conv_weight_key, conv_weight_value,
                  conv_bias_query, conv_bias_key, conv_bias_value, j_A_log,
                  j_dt_bias, q_loc, state_indices)

    new_conv_state = jnp.concatenate(
        [new_conv_state_q, new_conv_state_k, new_conv_state_v], axis=-1)

    return (new_conv_state, new_recurrent_state), output


def gdn_attention_core_tpu(
    mixed_qkv: torch.Tensor,
    b: torch.Tensor,
    a: torch.Tensor,
    core_attn_out: torch.Tensor,
    layer_name: str,
    mesh: jax.sharding.Mesh,
) -> None:
    """
    This acts as main bridge between PyTorch and JAX for the GDN core attention.
    Uses a robust, token-by-token scan to inherently handle any mix of
    ragged prefill and decode sequences without dynamic shape compilation errors.

    Some key details:
    1. Cache Mapping: We'll read vLLM's  `block_tables` and `query_start_loc`
       and translate them into static index arrays (`req_indices` and `state_indices`).
    2. JAX Scan: We use `jax.lax.scan` to perform a robust, token-by-token loop over
       the flat inputs. This allows us to handle ANY mix of prefill and decode tokens
       in a single compiled XLA graph.
    3. Conditional Updates: The `valid_mask` ensures that padded dummy tokens
       (used to keep the tensor shape static) do not corrupt the recurrent state
       in the cache.
    """
    fc = get_forward_context()
    attn_metadata = fc.attn_metadata[layer_name]

    layer_module = fc.no_compile_layers[layer_name]
    vllm_context = get_vllm_model_wrapper_context()

    n_kq = layer_module.num_k_heads
    n_v = layer_module.num_v_heads
    d_k = layer_module.head_k_dim
    d_v = layer_module.head_v_dim
    kernel_size = layer_module.conv_kernel_size

    j_mixed_qkv = jax_view(mixed_qkv)  # [num_tokens, dim]
    j_b = jax_view(b)
    j_a = jax_view(a)

    j_conv_weight = jax_view(layer_module.conv1d.weight)
    j_conv_bias = jax_view(layer_module.conv1d.bias
                           ) if layer_module.conv1d.bias is not None else None
    j_A_log = jax_view(layer_module.A_log)
    j_dt_bias = jax_view(layer_module.dt_bias)

    layer_idx = vllm_context.layer_name_to_kvcache_index[layer_name]
    conv_state, recurrent_state = vllm_context.kv_caches[layer_idx]

    # Map physical cache blocks
    flat_block_tables = jax_view(attn_metadata.block_tables)
    max_reqs = attn_metadata.seq_lens.shape[0]
    max_blocks_per_req = flat_block_tables.shape[0] // max_reqs
    block_tables_2d = jnp.reshape(flat_block_tables,
                                  (max_reqs, max_blocks_per_req))
    state_indices = block_tables_2d[:, 0].astype(jnp.int32)

    # Map tokens to their respective requests
    q_loc = jax_view(attn_metadata.query_start_loc)

    (new_conv_state,
     new_recurrent_state), j_output = run_jax_gdn_attention(j_mixed_qkv,
                                                            j_b,
                                                            j_a,
                                                            conv_state,
                                                            recurrent_state,
                                                            j_conv_weight,
                                                            j_conv_bias,
                                                            j_A_log,
                                                            j_dt_bias,
                                                            state_indices,
                                                            q_loc,
                                                            n_kq,
                                                            n_v,
                                                            d_k,
                                                            d_v,
                                                            kernel_size,
                                                            mesh=mesh)

    vllm_context.kv_caches[layer_idx] = (new_conv_state, new_recurrent_state)

    j_output_flat = j_output.reshape(core_attn_out.shape)
    core_attn_out.copy_(torch_view(j_output_flat))


def gdn_in_proj_tpu(
    hidden_states: torch.Tensor,
    qkvz_size: int,
    ba_size: int,
    prefix: str,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Torch OP replacement for gdn_in_proj.
    Executes the underlying linear layers (in_proj_qkvz and in_proj_ba) directly
    to mimic the old/non-kernel execution path that was reverted in:
    https://github.com/vllm-project/vllm/pull/36795

    Args:
        hidden_states: Tensor of shape (num_tokens, hidden_size)
        qkvz_size: int, unused but needed for signature
        ba_size: int, unused but needed for signature
        prefix: str
    Returns:
        mixed_qkvz: Tensor of shape (num_tokens, qkvz_size)
        ba: Tensor of shape (num_tokens, ba_size)
    """
    fc = get_forward_context()
    # The 'prefix' argument perfectly matches the key used to register the module
    layer_module = fc.no_compile_layers[prefix]

    # Run the original projections instead of the fused C++ kernel
    mixed_qkvz, _ = layer_module.in_proj_qkvz(hidden_states)
    ba, _ = layer_module.in_proj_ba(hidden_states)

    return mixed_qkvz, ba


def apply_gated_delta_net_torch_ops_patch(mesh: jax.sharding.Mesh) -> None:
    """
    This is a patch to inject the `gdn_attention_core` op so the
    Torch/GPU  kernel is bypassed in favor of the TPU kernel
    here:
    https://github.com/vllm-project/vllm/blob/697e4ff3528c72806a4d00ed9b7581332b9efd43/vllm/model_executor/models/qwen3_next.py#L671

    """
    try:
        import vllm.model_executor.models.qwen3_next  # noqa: F401
    except ImportError:
        pass

    # Ensure the op exists in the namespace, which initializes the OpOverloadPacket
    if hasattr(torch.ops, "vllm") and hasattr(torch.ops.vllm,
                                              "gdn_attention_core"):
        # dummy call to ensure the op is registered
        torch.ops.vllm.gdn_attention_core = functools.partial(
            gdn_attention_core_tpu, mesh=mesh)

    if hasattr(torch.ops.vllm, "gdn_in_proj"):
        # dummy call to ensure the op is registered
        torch.ops.vllm.gdn_in_proj = gdn_in_proj_tpu



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import jax.numpy as jnp
import pytest
import torch
from vllm.config import CacheConfig, CompilationMode, VllmConfig

from tpu_inference.platforms.tpu_platform import TpuPlatform


class TestTpuPlatform:

    @pytest.fixture
    def vllm_config(self):
        cache_config = CacheConfig(block_size=16,
                                   gpu_memory_utilization=0.9,
                                   cache_dtype="fp8")

        vllm_config = MagicMock(spec=VllmConfig)
        vllm_config.cache_config = cache_config
        vllm_config.model_config = MagicMock(dtype='bfloat16')
        vllm_config.scheduler_config = MagicMock(is_multimodal_model=False)
        vllm_config.parallel_config = MagicMock()
        vllm_config.compilation_config = MagicMock(mode="dynamo_trace_once",
                                                   backend="openxla")
        vllm_config.kv_transfer_config = None
        return vllm_config

    @pytest.mark.parametrize("chip_name,expected_dtype", [
        ("v6e", torch.float8_e5m2),
        ("v5e", torch.float8_e4m3fn),
    ])
    def test_fp8_dtype(self, chip_name, expected_dtype):
        mock_chip_type = MagicMock()
        mock_chip_type.name = chip_name

        with patch('tpu_inference.platforms.tpu_platform.init_logger'), \
             patch('tpu_info.device.get_local_chips', return_value=(mock_chip_type, None)), \
             patch('vllm.envs.VLLM_TPU_USING_PATHWAYS', False):
            assert TpuPlatform.fp8_dtype() == expected_dtype

    def test_basic_properties(self):
        assert not TpuPlatform.is_async_output_supported(True)
        assert not TpuPlatform.can_update_inplace()
        assert not TpuPlatform.is_pin_memory_available()
        assert TpuPlatform.use_all_gather()
        assert TpuPlatform.supports_v1(MagicMock())
        assert TpuPlatform.is_kv_cache_dtype_supported("fp8", MagicMock())
        assert TpuPlatform.use_sync_weight_loader()
        assert TpuPlatform.support_hybrid_kv_cache()
        assert TpuPlatform.inference_mode()
        assert TpuPlatform.get_lora_vocab_padding_size() == 1
        assert TpuPlatform.current_device().type == "cpu"

        expected_punica = "tpu_inference.lora.torch_punica_tpu.PunicaWrapperTPU"
        assert TpuPlatform.get_punica_wrapper() == expected_punica

        expected_communicator = "vllm.distributed.device_communicators.tpu_communicator.TpuCommunicator"
        assert TpuPlatform.get_device_communicator_cls(
        ) == expected_communicator

    def test_get_device_total_memory(self):
        with pytest.raises(NotImplementedError):
            TpuPlatform.get_device_total_memory()

    def test_get_infinity_values(self):
        min_val, max_val = TpuPlatform.get_infinity_values(jnp.float32)
        assert min_val == jnp.finfo(jnp.float32).min
        assert max_val == jnp.finfo(jnp.float32).max

    @patch('tpu_inference.platforms.tpu_platform.vllm_envs')
    def test_get_device_name_pathways(self, mock_envs):
        mock_envs.VLLM_TPU_USING_PATHWAYS = True
        assert TpuPlatform.get_device_name() == "TPU v6 lite"

    @patch('tpu_inference.platforms.tpu_platform.vllm_envs')
    def test_get_device_name_exception(self, mock_envs):
        mock_envs.VLLM_TPU_USING_PATHWAYS = False
        mock_device = MagicMock()
        mock_device.get_local_chips.side_effect = Exception("TPU Error")

        with patch.dict('sys.modules', {
                'tpu_info': MagicMock(),
                'tpu_info.device': mock_device
        }):
            assert TpuPlatform.get_device_name() == "TPU"

    def test_validate_request_random_seed(self):
        from vllm.sampling_params import SamplingParams, SamplingType
        params = MagicMock(spec=SamplingParams)
        params.sampling_type = SamplingType.RANDOM_SEED
        with pytest.raises(ValueError,
                           match="JAX does not support per-request seed"):
            TpuPlatform.validate_request(MagicMock(), params)

    def test_validate_request_valid(self):
        from vllm.sampling_params import SamplingParams, SamplingType
        params = MagicMock(spec=SamplingParams)
        params.sampling_type = SamplingType.GREEDY
        # Should execute cleanly without raising
        TpuPlatform.validate_request(MagicMock(), params)

    def test_get_attn_backend_cls_mla(self):
        with patch.dict('sys.modules',
                        {'tpu_inference.layers.vllm.backends': MagicMock()}):
            with patch(
                    "vllm.v1.attention.backends.registry.AttentionBackendEnum"
            ) as mock_attn_enum:
                mock_attn_enum.FLASH_ATTN_MLA = MagicMock()
                mock_attn_enum.FLASH_ATTN_MLA.get_path.return_value = "mla_path"

                config = MagicMock()
                config.use_mla = True

                res = TpuPlatform.get_attn_backend_cls(MagicMock(), config)
                assert res == "mla_path"

    def test_get_attn_backend_cls_fallback(self):
        with patch.dict('sys.modules',
                        {'tpu_inference.layers.vllm.backends': MagicMock()}):
            with patch(
                    "vllm.v1.attention.backends.registry.AttentionBackendEnum"
            ) as mock_attn_enum:
                mock_attn_enum.FLASH_ATTN = MagicMock()
                mock_attn_enum.FLASH_ATTN.get_path.return_value = "flash_path"

                config = MagicMock()
                config.use_mla = False

                selected = MagicMock()
                selected.name = "OTHER"

                res = TpuPlatform.get_attn_backend_cls(selected, config)
                assert res == "flash_path"

    @patch('tpu_inference.platforms.tpu_platform.vllm_envs')
    def test_check_and_update_config_pathways_assert(self, mock_envs,
                                                     vllm_config):
        mock_envs.VLLM_TPU_USING_PATHWAYS = True
        mock_envs.VLLM_ENABLE_V1_MULTIPROCESSING = True

        with pytest.raises(AssertionError,
                           match="VLLM_ENABLE_V1_MULTIPROCESSING must be 0"):
            TpuPlatform.check_and_update_config(vllm_config)

    @patch("tpu_inference.platforms.tpu_platform.envs.TPU_MULTIHOST_BACKEND",
           "")
    @patch("tpu_inference.platforms.tpu_platform.ShardingConfigManager")
    @patch(
        "tpu_inference.core.sched.dp_scheduler.update_vllm_config_for_dp_scheduler"
    )
    def test_check_and_update_config_single_host_uni(self, mock_update,
                                                     mock_sharding,
                                                     vllm_config):
        vllm_config.parallel_config.pipeline_parallel_size = 1
        vllm_config.scheduler_config.is_multimodal_model = True
        vllm_config.scheduler_config.disable_chunked_mm_input = False
        vllm_config.compilation_config.mode = "dummy"
        vllm_config.compilation_config.backend = ""
        vllm_config.cache_config = None  # Skip block size checks to isolate logic

        TpuPlatform.check_and_update_config(vllm_config)

        assert vllm_config.compilation_config.mode == CompilationMode.DYNAMO_TRACE_ONCE
        assert vllm_config.compilation_config.backend == "openxla"
        assert vllm_config.parallel_config.distributed_executor_backend == "uni"
        assert vllm_config.scheduler_config.disable_chunked_mm_input is True

        mock_sharding.from_vllm_config.assert_called_once_with(vllm_config)
        assert vllm_config.sharding_config == mock_sharding.from_vllm_config.return_value
        mock_update.assert_called_once_with(vllm_config)

    @patch("tpu_inference.platforms.tpu_platform.envs.TPU_MULTIHOST_BACKEND",
           "")
    @patch("tpu_inference.platforms.tpu_platform.ShardingConfigManager")
    @patch(
        "tpu_inference.core.sched.dp_scheduler.update_vllm_config_for_dp_scheduler"
    )
    def test_check_and_update_config_single_host_multiproc(
            self, mock_update, mock_sharding, vllm_config):
        vllm_config.parallel_config.pipeline_parallel_size = 2
        vllm_config.cache_config = None

        with patch.dict(
                'sys.modules',
            {'tpu_inference.executors.multiproc_executor': MagicMock()}):
            from tpu_inference.executors.multiproc_executor import \
                MultiprocExecutor
            TpuPlatform.check_and_update_config(vllm_config)
            assert vllm_config.parallel_config.distributed_executor_backend == MultiprocExecutor

    @patch("tpu_inference.platforms.tpu_platform.envs.TPU_MULTIHOST_BACKEND",
           "ray")
    @patch("tpu_inference.platforms.tpu_platform.ShardingConfigManager")
    @patch(
        "tpu_inference.core.sched.dp_scheduler.update_vllm_config_for_dp_scheduler"
    )
    def test_check_and_update_config_ray(self, mock_update, mock_sharding,
                                         vllm_config):
        vllm_config.cache_config = None

        with patch.dict(
                'sys.modules',
            {'tpu_inference.executors.ray_distributed_executor': MagicMock()}):
            from tpu_inference.executors.ray_distributed_executor import \
                RayDistributedExecutor
            TpuPlatform.check_and_update_config(vllm_config)
            assert vllm_config.parallel_config.distributed_executor_backend == RayDistributedExecutor

    @patch("tpu_inference.platforms.tpu_platform.envs.TPU_MULTIHOST_BACKEND",
           "unknown_backend")
    @patch("tpu_inference.platforms.tpu_platform.ShardingConfigManager")
    @patch(
        "tpu_inference.core.sched.dp_scheduler.update_vllm_config_for_dp_scheduler"
    )
    def test_check_and_update_config_unknown(self, mock_update, mock_sharding,
                                             vllm_config):
        vllm_config.cache_config = None
        TpuPlatform.check_and_update_config(vllm_config)

        # Fallback behaviour defaults to `uni`
        assert vllm_config.parallel_config.distributed_executor_backend == "uni"

    @patch("tpu_inference.platforms.tpu_platform.envs.TPU_MULTIHOST_BACKEND",
           "")
    @patch("tpu_inference.platforms.tpu_platform.ShardingConfigManager")
    @patch(
        "tpu_inference.core.sched.dp_scheduler.update_vllm_config_for_dp_scheduler"
    )
    @patch("tpu_inference.platforms.tpu_platform.logger")
    def test_check_and_update_config_block_size(self, mock_logger, mock_update,
                                                mock_sharding, vllm_config):
        mock_pallas = MagicMock()
        mock_pallas.get_page_size.return_value = 16
        mock_pallas.get_min_page_size.return_value = 64

        vllm_config.cache_config = MagicMock()
        vllm_config.cache_config.user_specified_block_size = False
        vllm_config.cache_config.block_size = 16
        vllm_config.model_config.use_mla = False

        with patch.dict(
                'sys.modules', {
                    'tpu_inference.layers.vllm.backends.flash_attn':
                    MagicMock(PallasAttentionBackend=mock_pallas)
                }):
            TpuPlatform.check_and_update_config(vllm_config)
            # The test should check and update min_page_size and override block_size if min_page_size is larger
            assert vllm_config.cache_config.block_size == 64

    def test_update_block_size_for_backend(self, vllm_config):
        # Current implementation explicitly performs 'pass'. Ensure it returns safely.
        TpuPlatform.update_block_size_for_backend(vllm_config)



# SPDX-License-Identifier: Apache-2.0
"""Quantized matmul kernel."""

import functools

import jax
import jax.numpy as jnp
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

from tpu_inference.kernels.quantized_matmul import util
from tpu_inference.kernels.quantized_matmul.tuned_block_sizes import (
    TunedValue, get_device_vmem_limit, get_tuned_block_sizes)
from tpu_inference.kernels.quantized_matmul.util import (get_kernel_name,
                                                         next_multiple,
                                                         unfold_args)

quantize_tensor = util.quantize_tensor


def matmul_kernel(
    x_ref: jax.Array,  # (batch_block_size, in_block_size)
    w_q_ref: jax.Array,  # (out_block_size, in_block_size)
    w_scale_ref: jax.Array,  # (1, out_block_size)
    x_abs_max_ref: jax.Array,  # (1, batch_block_size)
    out_ref: jax.Array,  # (batch_block_size, out_block_size)
    acc_scratch: jax.Array,  # (batch_block_size, out_block_size)
    x_q_scratch: jax.Array,  # (batch_block_size, in_block_size)
    x_scale_scratch: jax.Array,  # (batch_block_size, 1)
    *,
    x_q_dtype: jnp.dtype,
    save_acc: bool,
    save_x_q: bool,
):
    out_idx, in_idx = pl.program_id(1), pl.program_id(2)
    n_in = pl.num_programs(2)
    x_ref_dtype = x_ref.dtype

    quantize_activation = x_q_dtype != x_ref_dtype

    # Initialize conditional logic.
    if save_x_q:
        assert quantize_activation
        assert x_q_scratch is not None
        assert x_scale_scratch is not None
        quant = out_idx == 0
    else:
        assert x_q_scratch is None
        assert x_scale_scratch is None
        quant = quantize_activation

    if save_acc:
        assert acc_scratch is not None
        is_first_step = in_idx == 0
        is_last_step = in_idx == (n_in - 1)
    else:
        assert acc_scratch is None
        is_first_step = True
        is_last_step = True

    acc_dtype = jnp.float32
    if quantize_activation and jnp.issubdtype(w_q_ref.dtype, jnp.integer):
        acc_dtype = jnp.int32

    # Start of actual computation logic.
    def matmul_body(quant: bool, is_first_step: bool, is_last_step: bool):
        if quantize_activation:
            if quant:
                x_q_tmp, x_scale_tmp = util.quantize_array(
                    x_ref[...],
                    x_abs_max_ref[...],
                    x_q_dtype,
                )

                if save_x_q:
                    x_q_scratch[...] = x_q_tmp
                    x_scale_scratch[...] = x_scale_tmp

            else:
                assert save_x_q
                x_q_tmp = x_q_scratch[...]
                if is_last_step:
                    x_scale_tmp = x_scale_scratch[...]

            acc = jax.lax.dot_general(
                x_q_tmp,
                w_q_ref[...],
                (((1, ), (1, )), ((), ())),
                preferred_element_type=acc_dtype,
            )
        else:
            acc = jax.lax.dot_general(
                x_ref[...],
                w_q_ref[...],
                (((1, ), (1, )), ((), ())),
                preferred_element_type=acc_dtype,
            )

        if not is_first_step:
            acc += acc_scratch[...]

        if is_last_step:
            acc *= w_scale_ref[...]
            if quantize_activation:
                # TODO(kyuyeunk): Investigate caching broadcast.
                acc *= x_scale_tmp
            out_ref[...] = acc.astype(x_ref_dtype)
        else:
            assert save_acc
            acc_scratch[...] = acc

    unfold_args((quant, is_first_step, is_last_step), (), matmul_body)


@jax.jit(static_argnames=[
    "x_q_dtype",
    "tuned_value",
])
def quantized_matmul_kernel(
    x: jax.Array,  # [bs, n_in]
    w_q: jax.Array,  # [n_out, n_in]
    w_scale: jax.Array,  # [n_out]
    w_zp: jax.Array | None = None,  # [n_out]
    block_size: int | None = None,
    x_q_dtype: jnp.dtype | None = None,
    *,
    tuned_value: TunedValue | None = None,
) -> jax.Array:
    """Quantized matmul kernel.

    Args:
      x: Input unquantized array.
      w_q: Weight quantized array. [n_output_features, n_input_features]
      w_scale: Weight quantization scale. [n_output_features]
      w_zp: Weight zero point for asymmetric quantization.
      block_size: Block size for subchannel quantization.
      x_q_dtype: Quantization type of the input. If None or if the value is the
        same as x.dtype, then no quantization is applied.
      tuned_value: Kernel tuned values for optimal performance.

    Returns:
      Quantized matmul result.
    """

    if w_zp is not None:
        raise NotImplementedError("zero_point is not supported.")
    if block_size is not None:
        raise NotImplementedError("block_size is not supported.")

    if x_q_dtype is None:
        x_q_dtype = x.dtype
    quantize_activation = x_q_dtype != x.dtype

    # Pallas kernel only has access to a single block of the input. Therefere,
    # for per-token quantization, abs max has to be computed outside of the
    # kernel.
    x_abs_max = jnp.max(jnp.abs(x), axis=-1, keepdims=False)  # [bs]
    # Pallas requires minormost dim to be a multiple of sublane size 128.
    # Therefore, instead of using [bs, 1], we reshape this into [1, bs]
    x_abs_max = jnp.expand_dims(x_abs_max, axis=0)  # [1, bs]
    assert x_abs_max.shape == (1, x.shape[0])

    orig_n_batch, orig_n_in = x.shape
    orig_n_out, _ = w_q.shape

    if tuned_value is None:
        tuned_value = get_tuned_block_sizes(
            n_batch=orig_n_batch,
            n_out=orig_n_out,
            n_in=orig_n_in,
            x_q_dtype=jnp.dtype(x_q_dtype).name,
            w_q_dtype=jnp.dtype(w_q.dtype).name,
        )
    batch_block_size = tuned_value.batch_block_size
    out_block_size = tuned_value.out_block_size
    in_block_size = tuned_value.in_block_size

    # Pad the inputs to be multiple of block size.
    padded_n_batch = next_multiple(orig_n_batch, batch_block_size)
    if orig_n_batch < padded_n_batch:
        x = jnp.pad(x, ((0, padded_n_batch - orig_n_batch), (0, 0)))
        x_abs_max = jnp.pad(x_abs_max,
                            ((0, 0), (0, padded_n_batch - orig_n_batch)))
    padded_n_out = next_multiple(orig_n_out, out_block_size)
    if orig_n_out < padded_n_out:
        w_q = jnp.pad(w_q, ((0, padded_n_out - orig_n_out), (0, 0)))
        w_scale = jnp.pad(w_scale, (0, padded_n_out - orig_n_out))
    padded_n_in = next_multiple(orig_n_in, in_block_size)
    if orig_n_in < padded_n_in:
        x = jnp.pad(x, ((0, 0), (0, padded_n_in - orig_n_in)))
        w_q = jnp.pad(w_q, ((0, 0), (0, padded_n_in - orig_n_in)))

    if w_scale.dtype != jnp.float32:
        w_scale = w_scale.astype(jnp.float32)
    w_scale = jnp.expand_dims(w_scale, axis=0)  # [1, n_output_features]

    n_batch = padded_n_batch // batch_block_size
    n_out = padded_n_out // out_block_size
    n_in = padded_n_in // in_block_size

    save_acc = n_in > 1
    # Remove redundant input quantization logic by caching quantized input. For
    # best performance, only enable this behavior when single input block is
    # used per batch.
    save_x_q = quantize_activation and n_in == 1 and n_out > 1

    acc_dtype = jnp.float32
    if quantize_activation and jnp.issubdtype(w_q.dtype, jnp.integer):
        acc_dtype = jnp.int32

    vmem_limit_bytes = util.get_vmem_limit(
        n_batch=n_batch,
        n_out=n_out,
        n_in=n_in,
        batch_block_size=batch_block_size,
        out_block_size=out_block_size,
        in_block_size=in_block_size,
        x_dtype=x.dtype,
        x_q_dtype=x_q_dtype,
        w_q_dtype=w_q.dtype,
        scale_dtype=jnp.float32,
        out_dtype=x.dtype,
        acc_dtype=acc_dtype,
        save_acc=save_acc,
        save_x_q=save_x_q,
        upper_limit_bytes=get_device_vmem_limit(),
    )

    kernel = pl.pallas_call(
        functools.partial(
            matmul_kernel,
            x_q_dtype=x_q_dtype,
            save_acc=save_acc,
            save_x_q=save_x_q,
        ),
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=0,
            in_specs=[
                pl.BlockSpec((batch_block_size, in_block_size), lambda b, o, i:
                             (b, i)),  # x
                pl.BlockSpec((out_block_size, in_block_size), lambda b, o, i:
                             (o, i)),  # w_q
                pl.BlockSpec((1, out_block_size), lambda b, o, i:
                             (0, o)),  # w_scale
                pl.BlockSpec((1, batch_block_size), lambda b, o, i:
                             (0, b)),  # x_abs_max
            ],
            out_specs=pl.BlockSpec((batch_block_size, out_block_size),
                                   lambda b, o, i: (b, o)),
            scratch_shapes=[
                (pltpu.VMEM((batch_block_size, out_block_size), acc_dtype)
                 if save_acc else None),  # acc_scratch
                (pltpu.VMEM((batch_block_size, in_block_size), x_q_dtype)
                 if save_x_q else None),  # x_q_scratch
                (pltpu.VMEM((batch_block_size, 1), jnp.float32)
                 if save_x_q else None),  # x_scale_scratch
            ],
            grid=(n_batch, n_out, n_in),
        ),
        out_shape=jax.ShapeDtypeStruct((padded_n_batch, padded_n_out),
                                       x.dtype),
        compiler_params=pltpu.CompilerParams(
            dimension_semantics=("parallel", "arbitrary", "arbitrary"),
            vmem_limit_bytes=vmem_limit_bytes,
        ),
    )

    util.validate_inputs(
        x=x,
        w_q=w_q,
        w_scale=w_scale,
        x_abs_max=x_abs_max,
        batch_block_size=batch_block_size,
        out_block_size=out_block_size,
        in_block_size=in_block_size,
    )

    # The named_scope is used for autotune.
    kernel_name = get_kernel_name(tuned_value)
    with jax.named_scope(kernel_name):
        out = kernel(x, w_q, w_scale, x_abs_max)

    return out[:orig_n_batch, :orig_n_out]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import field
from types import SimpleNamespace
from typing import Any, Tuple
from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from flax.typing import PRNGKey
from jax.sharding import Mesh
from vllm.config import ModelConfig

from tpu_inference.models.jax.llama_guard_4 import (LlamaGuard4ForCausalLM,
                                                    LlamaGuard4WeightLoader)


class MockParamLlamaGuard4:
    """A mock for a parameter used in the LlamaGuard4 model."""
    shape: Tuple[int, ...]
    dtype: jnp.dtype = jnp.bfloat16
    sharding_spec: Tuple[str | None, ...] | None = None
    value: Any = field(init=False)
    out_sharding: Any = field(init=False)

    def __init__(self, shape=(32, 128)):
        self.shape = shape
        self.value = jnp.zeros(self.shape, dtype=self.dtype)
        # The sharding spec is accessed during weight loading
        self.out_sharding = SimpleNamespace(spec=self.sharding_spec)

    # Allow the mock parameter's value to be updated
    def __setattr__(self, name, value):
        if name in [
                'value', 'shape', 'dtype', 'out_sharding', 'sharding_spec'
        ]:
            self.__dict__[name] = value
        else:
            super().__setattr__(name, value)


class MockVllmConfig:
    """A mock VllmConfig sufficient for testing the LlamaGuard4 model."""

    def __init__(self,
                 model_name: str,
                 random_weights: bool = False,
                 tensor_parallelism: int = 1):
        self.model_config = MagicMock(spec=ModelConfig)
        self.load_config = MagicMock()
        self.load_config.download_dir = None

        # Downsizing the following to avoid OOM
        self.model_config.get_vocab_size.return_value = 1024
        self.model_config.get_hidden_size.return_value = 128
        self.model_config.model = model_name

        self.additional_config = {
            "random_weights": random_weights,
            "sharding": {
                "sharding_strategy": {
                    "tensor_parallelism": tensor_parallelism
                }
            }
        }

        self.cache_config = MagicMock(cache_dtype="auto")

        # Mock the underlying HF config values for parameter detection
        # Downsized to avoid OOM
        text_config_mock = MagicMock()
        text_config_mock.num_attention_heads = 4
        text_config_mock.num_key_value_heads = 2
        text_config_mock.head_dim = 32

        hf_config_mock = MagicMock()
        hf_config_mock.text_config = text_config_mock

        vision_config_mock = MagicMock()
        vision_config_mock.image_size = 336
        vision_config_mock.patch_size = 14
        vision_config_mock.hidden_size = 1408
        vision_config_mock.num_attention_heads = 16
        vision_config_mock.rope_theta = 10000.0
        vision_config_mock.intermediate_size = 5632
        vision_config_mock.projector_input_dim = 4096
        vision_config_mock.projector_output_dim = 4096
        vision_config_mock.projector_dropout = 0.0
        hf_config_mock.vision_config = vision_config_mock
        hf_config_mock.image_token_index = 200092

        self.model_config.hf_config = hf_config_mock


@pytest.fixture(scope="module")
def mesh():
    """
    Creates a mesh with all required axes for testing.
    """
    if not jax.devices():
        pytest.skip("No JAX devices available for mesh creation.")

    devices = np.array(jax.local_devices())

    num_devices = len(devices)
    device_mesh = devices.reshape((num_devices, 1, 1, 1))

    with Mesh(device_mesh,
              axis_names=('data', 'attn_dp', 'model', 'expert')) as m:
        yield m


@pytest.fixture
def rng() -> PRNGKey:
    """Provides a reusable JAX PRNGKey."""
    return jax.random.PRNGKey(42)


@pytest.fixture
def mock_vllm_config_llama_guard_4() -> MockVllmConfig:
    return MockVllmConfig(model_name="meta-llama/Llama-Guard-4-12B")


@pytest.fixture(autouse=True)
def mock_get_pp_group():
    with patch("tpu_inference.models.jax.llama_guard_4.get_pp_group",
               return_value=MagicMock(is_first_rank=True,
                                      is_last_rank=True,
                                      rank_in_group=0,
                                      world_size=1)):
        yield


class TestLlamaGuard4ForCausalLM:
    """Tests for the main LlamaGuard4ForCausalLM model class."""

    def test_init_llama_guard_4(self, mock_vllm_config_llama_guard_4, rng,
                                mesh):
        """Tests correct initialization and parameter detection."""
        with jax.set_mesh(mesh):
            model = LlamaGuard4ForCausalLM(mock_vllm_config_llama_guard_4, rng,
                                           mesh)

        # Check model name is correctly set in the config
        assert "llama-guard-4" in model.vllm_config.model_config.model.lower()

        assert model.hidden_size == 128

    def test_create_model_with_random_weights(self,
                                              mock_vllm_config_llama_guard_4,
                                              rng, mesh):
        """
        Tests that random weight initialization creates concrete, non-zero-variance arrays.
        """
        with jax.set_mesh(mesh):
            model = LlamaGuard4ForCausalLM(
                vllm_config=mock_vllm_config_llama_guard_4,
                rng=rng,
                mesh=mesh,
                force_random_weights=True)

            embedding_weight = model.embedder.input_embedding_table_VD.value
            attention_q_kernel = model.layers[0].attn.kernel_q_proj_DNH.value
            final_norm_scale = model.final_norm.scale.value

            assert isinstance(embedding_weight, jax.Array)
            assert isinstance(attention_q_kernel, jax.Array)
            assert isinstance(final_norm_scale, jax.Array)

            assert jnp.std(embedding_weight) > 0
            assert jnp.std(attention_q_kernel) > 0

            assert jnp.all(final_norm_scale == 1.0)

    def test_load_weights_called_correctly(self, rng, mesh):
        """Tests that the weight loader is called correctly for checkpoint loading."""
        vllm_config = MockVllmConfig(model_name="llama-guard-4-test",
                                     random_weights=False)
        with jax.set_mesh(mesh):
            model = LlamaGuard4ForCausalLM(vllm_config, rng, mesh)

        with patch(
                'tpu_inference.models.jax.llama_guard_4.LlamaGuard4WeightLoader'
        ) as mock_loader_cls:
            mock_loader_instance = MagicMock()
            mock_loader_cls.return_value = mock_loader_instance
            model.load_weights(rng)

            mock_loader_cls.assert_called_once_with(vllm_config=vllm_config,
                                                    hidden_size=128,
                                                    attn_heads=4,
                                                    num_key_value_heads=2,
                                                    attn_head_dim=32)
            mock_loader_instance.load_weights.assert_called_once_with(model)


class TestLlamaGuard4WeightLoader:
    """Tests for the LlamaGuard4WeightLoader class."""

    @pytest.fixture
    def weight_loader(self):
        return LlamaGuard4WeightLoader(
            vllm_config=MockVllmConfig("test-model"),
            hidden_size=5120,
            attn_heads=40,
            num_key_value_heads=8,
            attn_head_dim=128)

    @pytest.mark.parametrize("hf_key, expected", [
        ("language_model.model.layers.15.self_attn.q_proj.weight",
         "layers.15.attn.kernel_q_proj_DNH"),
        ("language_model.model.layers.0.feed_forward.gate_proj.weight",
         "layers.0.custom_module.kernel_gating_DF"),
        ("language_model.model.embed_tokens.weight",
         "embedder.input_embedding_table_VD"),
        ("language_model.model.norm.weight", "final_norm.scale"),
        ("language_model.lm_head.weight", "lm_head.input_embedding_table_DV"),
        ("unmapped.key.name", "unmapped.key.name"),
    ])
    def test_map_loaded_to_standardized_name(self, weight_loader, hf_key,
                                             expected):
        """Tests the mapping from HuggingFace key names to internal names."""
        assert weight_loader.map_loaded_to_standardized_name(
            hf_key) == expected

    def test_load_weights_transformation(self, weight_loader, rng, mesh):
        """Tests that weights are correctly reshaped, transposed, and loaded."""
        vllm_config = MockVllmConfig(model_name="llama-guard-4-small-test",
                                     random_weights=False)

        with jax.set_mesh(mesh):
            model = LlamaGuard4ForCausalLM(vllm_config, rng, mesh)

        hidden_size = 5120
        vocab_size = 202048

        original_weight = jnp.ones((vocab_size, hidden_size))
        dummy_weights = [
            ("language_model.model.embed_tokens.weight", original_weight),
        ]
        weight_loader.names_and_weights_generator = dummy_weights

        # Mock get_param to return a mock param with the target shape
        mock_param = MockParamLlamaGuard4(shape=(vocab_size, hidden_size))

        with patch("tpu_inference.models.jax.llama_guard_4.get_param", return_value=mock_param), \
            patch("tpu_inference.models.jax.llama_guard_4.shard_put", return_value=jnp.ones(mock_param.value.shape)) as mock_shard_put:

            weight_loader.load_weights(model)

            # Assert that shard_put was called with the correctly transposed weight
            mock_shard_put.assert_called_once()

            # Get the actual array passed to shard_put
            called_with_weight = mock_shard_put.call_args[0][0]

            # Check if the shape of the array passed to shard_put matches the model's expected shape.
            assert called_with_weight.shape == mock_param.value.shape



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from dataclasses import dataclass, field
from typing import Any

import jax


@functools.partial(
    jax.tree_util.register_dataclass,
    data_fields=[
        "input_positions",
        "block_tables",
        "seq_lens",
        "query_start_loc",
        "request_distribution",
    ],
    meta_fields=[],
    drop_fields=["query_start_loc_cpu", "seq_lens_cpu"],
)
@dataclass
class AttentionMetadata(object):
    # (padded_total_num_scheduled_tokens,)
    input_positions: jax.Array
    # (max_num_seqs * max_num_blocks_per_req,)
    # None for pooling models that using no KV cache
    block_tables: jax.Array | None = None
    # (max_num_seqs,)
    seq_lens: jax.Array = None
    # (max_num_seqs + 1,)
    query_start_loc: jax.Array = None
    # (3,)
    request_distribution: jax.Array = None

    query_start_loc_cpu: Any = field(init=False)
    seq_lens_cpu: Any = field(init=False)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from functools import partial
from typing import (Callable, List, Literal, NamedTuple, Optional, TypedDict,
                    Union)

import jax
import jax.numpy as jnp
import numpy as np
from flax import nnx
from jax.sharding import Mesh
from transformers.models.qwen2_5_vl.configuration_qwen2_5_vl import (
    Qwen2_5_VLConfig, Qwen2_5_VLVisionConfig)
from vllm.config import VllmConfig

from tpu_inference import utils as utils
from tpu_inference.distributed.jax_parallel_state import get_pp_group
from tpu_inference.layers.common.attention_interface import \
    sharded_flash_attention
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.jax.layers import FlaxUtils
from tpu_inference.layers.jax.linear import JaxEinsum
from tpu_inference.layers.jax.pp_utils import PPMissingLayer
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.jax_intermediate_tensor import \
    JaxIntermediateTensors
from tpu_inference.models.jax.qwen2 import Qwen2Model
# from vllm.model_executor.models.interfaces import MultiModalEmbeddings
from tpu_inference.models.jax.utils.multi_modal_utils import (
    MultiModalEmbeddings, merge_multimodal_embeddings)
from tpu_inference.models.jax.utils.weight_utils import StandardWeightLoader

logger = init_logger(__name__)

init_fn = nnx.initializers.uniform()
modeling_flax_utils = FlaxUtils()

DEFAULT_BLOCK_K_MAJOR = 128


class SegmentIds(NamedTuple):
    """SegmentIds for Q and KV sequences.

  SegmentIds are used to generate segment mask, which prevents attention between
  different segments in the input sequence. Each array is a list of ids
  (integers).
  Only the token with the same id can attend to each other.

  Attributes:
    q: segment ids along the Q sequence.
    kv: segment ids along the KV sequence.
  """

    q: jax.Array  # [batch_size, q_seq_len]
    kv: jax.Array  # [batch_size, kv_seq_len]


class Qwen2_5_VLImagePixelInputs(TypedDict):
    type: Literal["pixel_values"]
    pixel_values: jax.Array
    """Shape:
    `(num_patches, num_channels * patch_size * patch_size)`
    """

    image_grid_thw: tuple[tuple[int, int, int], ...]
    """Shape: `(num_images, 3)`
    This should be in `(grid_t, grid_h, grid_w)` format.
    """


# NOTE: We are not supporting embedding inputs for now
# The code here makes the struture consistent and
# makes iteasier for future implementation
class Qwen2_5_VLImageEmbeddingInputs(TypedDict):
    type: Literal["image_embeds"]
    image_embeds: jax.Array
    """Supported types:
    - list[`jax.Array`]: A list of tensors holding all images' features.
        Each tensor holds an image's features.
    - `jax.Array`: A tensor holding all images' features (concatenation of
        all images' feature tensors).

    Tensor shape: `(num_image_features, hidden_size)`
    - `num_image_features` varies based on
        the number and resolution of the images.
    - `hidden_size` must match the hidden size of language model backbone.
    """

    image_grid_thw: jax.Array
    """Shape: `(num_images, 3)`
    This should be in `(grid_t, grid_h, grid_w)` format.
    """


Qwen2_5_VLImageInputs = Union[Qwen2_5_VLImagePixelInputs,
                              Qwen2_5_VLImageEmbeddingInputs]


class Qwen2_5_VisionMLP(nnx.Module):

    def __init__(self, config: Qwen2_5_VLVisionConfig, dtype: jnp.dtype,
                 rngs: nnx.Rngs):
        in_features = config.hidden_size
        hidden_features = config.intermediate_size
        act_fn = modeling_flax_utils.ACT2FN[config.hidden_act]
        self.gate_proj = nnx.Linear(
            in_features,
            hidden_features,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, (None, "model")),
            bias_init=nnx.with_partitioning(init_fn, ("model", )),
            rngs=rngs,
        )
        self.up_proj = nnx.Linear(
            in_features,
            hidden_features,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, (None, "model")),
            bias_init=nnx.with_partitioning(init_fn, ("model", )),
            rngs=rngs,
        )
        self.down_proj = nnx.Linear(
            hidden_features,
            in_features,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, ("model", None)),
            bias_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rngs,
        )
        self.act_fn = act_fn

    def __call__(self, x: jax.Array) -> jax.Array:
        gate = self.act_fn(self.gate_proj(x))
        up = self.up_proj(x)
        fuse = gate * up
        result = self.down_proj(fuse)
        return result


def apply_rotary_pos_emb_vision(x: jax.Array,
                                rotary_pos_emb: jax.Array) -> jax.Array:
    # x: [B, T, N, H]
    # rotary_pos_emb: [T, H//2]
    _, _, _, H = x.shape
    half_dim = H // 2

    # [B, T, N, H//2]
    x_real = x[..., :half_dim]
    x_imag = x[..., half_dim:]

    # [T, H//2]
    cos_emb = jnp.cos(rotary_pos_emb)
    sin_emb = jnp.sin(rotary_pos_emb)

    # [1, T, 1, H//2]
    cos_emb = cos_emb[None, :, None, :]
    sin_emb = sin_emb[None, :, None, :]

    # [B, T, N, H//2]
    x_rotated_real = x_real * cos_emb - x_imag * sin_emb
    x_rotated_imag = x_real * sin_emb + x_imag * cos_emb

    # [B, T, N, H]
    x_rotated = jnp.concatenate([x_rotated_real, x_rotated_imag], axis=-1)

    return x_rotated


def generate_window_segment_ids(cu_seqlens: jax.Array, seq_len: int,
                                padded_seq_len: int) -> SegmentIds:
    """Generates segment IDs for windowed attention

    Args:
        cu_seqlens: A 1D array of cumulative sequence lengths for each window.
            e.g., [0, len_win0, len_win0+len_win1, ...]

    Returns:
        A SegmentIds object for flash_attention.
    """
    indices = jnp.arange(seq_len, dtype=jnp.int32)
    segment_ids = jnp.searchsorted(cu_seqlens[1:], indices, side='right') + 1
    padding_segment_ids = jnp.zeros(padded_seq_len - seq_len, dtype=jnp.int32)
    segment_ids = jnp.concatenate([segment_ids, padding_segment_ids])
    segment_ids = segment_ids.reshape(1, -1)

    return SegmentIds(q=segment_ids, kv=segment_ids)


class Qwen2_5_VisionAttention(nnx.Module):

    def __init__(self, config: Qwen2_5_VLConfig, dtype: jnp.dtype,
                 rngs: nnx.Rngs, mesh: Mesh):
        vision_config = config.vision_config
        self.hidden_size = vision_config.hidden_size
        self.num_heads = vision_config.num_heads
        self.num_kv_heads = self.num_heads
        self.rope_theta = config.rope_theta
        self.rope_scaling = getattr(config, "rope_scaling", None)
        self.head_dim_original = self.hidden_size // self.num_heads

        sharding_size = mesh.shape["model"]
        self.num_heads = utils.get_padded_num_heads(self.num_heads,
                                                    sharding_size)
        self.num_kv_heads = utils.get_padded_num_heads(self.num_kv_heads,
                                                       sharding_size)
        self.head_dim = utils.get_padded_head_dim(self.head_dim_original)

        # TODO: Wenlong: Do not consider padding for now
        self.head_dim = self.head_dim_original

        self.mesh = mesh

        self.qkv_proj = nnx.Linear(
            self.hidden_size,
            3 * self.hidden_size,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, (None, "model")),
            bias_init=nnx.with_partitioning(init_fn, ("model", )),
            rngs=rngs,
        )

        self.proj = nnx.Linear(
            self.hidden_size,
            self.hidden_size,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, ("model", None)),
            bias_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rngs,
        )
        self.flash_attention = sharded_flash_attention(
            mesh=mesh,
            causal=False,
            sm_scale=1.0 / math.sqrt(self.head_dim),
            vmem_limit_bytes=128 * 1024 * 1024,
        )

    def __call__(
        self,
        x: jax.Array,
        rotary_pos_emb: jax.Array,
        cu_window_seqlens: Optional[jax.Array] = None,
        use_fullattn: bool = True,
    ) -> jax.Array:
        T, B, D = x.shape
        assert B == 1, "Vision attention currently only supports batch size 1"
        # [T, B, D] -> [T, B, 3 * D]
        qkv = self.qkv_proj(x)

        # Split into Q, K, V.
        # NOTE: simplified from vLLM's split_qkv,
        # may need to revisit for tp>1
        # [T, B, 3 * D] -> 3 *[T, B, D]
        q, k, v = jnp.split(qkv, 3, axis=-1)

        # [T, B, N, H]
        q = q.reshape(T, B, self.num_heads, self.head_dim)
        k = k.reshape(T, B, self.num_heads, self.head_dim)
        v = v.reshape(T, B, self.num_heads, self.head_dim)

        # [T, B, N, H] -> [B, T, N, H]
        q = jnp.transpose(q, (1, 0, 2, 3))
        k = jnp.transpose(k, (1, 0, 2, 3))
        v = jnp.transpose(v, (1, 0, 2, 3))

        # rotary_pos_emb shape: (T, H)
        q = apply_rotary_pos_emb_vision(q, rotary_pos_emb)
        k = apply_rotary_pos_emb_vision(k, rotary_pos_emb)

        # NOTE: an extra transpose because we need to
        # align the correctness with vLLM's design.
        # Might be able to remove one once implemented.
        # [B, T, N, H] -> [B, N, T, H]
        q = jnp.transpose(q, (0, 2, 1, 3))
        k = jnp.transpose(k, (0, 2, 1, 3))
        v = jnp.transpose(v, (0, 2, 1, 3))

        # Pad the sequence length to be a multiple of 128 for flash_attention
        block_k_major = DEFAULT_BLOCK_K_MAJOR
        T_attn = q.shape[2]
        padded_T = (T_attn + block_k_major -
                    1) // block_k_major * block_k_major
        pad_width = ((0, 0), (0, 0), (0, padded_T - T_attn), (0, 0))

        q = jnp.pad(q, pad_width, 'constant')
        k = jnp.pad(k, pad_width, 'constant')
        v = jnp.pad(v, pad_width, 'constant')

        segment_ids = generate_window_segment_ids(cu_window_seqlens, T_attn,
                                                  padded_T)

        # TODO (jacobplatin): add support for quantized KV cache?
        output = self.flash_attention(q, k, v, segment_ids)

        # Unpad the output
        output = output[:, :, :T_attn, :]

        # [B, N, T, H] -> [T, B, N, H]
        output = jnp.transpose(output, (2, 0, 1, 3))

        output = output.reshape(T, B, D)

        output = self.proj(output)

        return output


class Qwen2_5_VisionBlock(nnx.Module):

    def __init__(self, config: Qwen2_5_VLConfig, dtype: jnp.dtype,
                 rngs: nnx.Rngs, mesh: Mesh):
        vision_config = config.vision_config
        dim = vision_config.hidden_size
        norm_layer = partial(nnx.RMSNorm,
                             epsilon=config.rms_norm_eps,
                             param_dtype=dtype,
                             scale_init=nnx.with_partitioning(
                                 init_fn, (None, )))

        self.norm1 = norm_layer(dim, dtype=dtype, rngs=rngs)
        self.norm2 = norm_layer(dim, dtype=dtype, rngs=rngs)
        self.attn = Qwen2_5_VisionAttention(config=config,
                                            dtype=dtype,
                                            rngs=rngs,
                                            mesh=mesh)
        self.mlp = Qwen2_5_VisionMLP(config=vision_config,
                                     dtype=dtype,
                                     rngs=rngs)

    def __call__(self,
                 x: jax.Array,
                 rotary_pos_emb: jax.Array,
                 cu_window_seqlens: Optional[jax.Array] = None,
                 use_fullattn: bool = True) -> jax.Array:

        x = x + self.attn(self.norm1(x), rotary_pos_emb, cu_window_seqlens,
                          use_fullattn)
        x = x + self.mlp(self.norm2(x))

        return x


class Qwen2_5_VisionPatchEmbed(nnx.Module):

    def __init__(
        self,
        rngs: nnx.Rngs,
        patch_size: int = 14,
        temporal_patch_size: int = 2,
        in_channels: int = 3,
        hidden_size: int = 1152,
        dtype: jnp.dtype = jnp.bfloat16,
    ) -> None:
        self.patch_size = patch_size
        self.temporal_patch_size = temporal_patch_size
        self.hidden_size = hidden_size
        kernel_size = (temporal_patch_size, patch_size, patch_size)
        self.proj = nnx.Conv(in_features=in_channels,
                             out_features=hidden_size,
                             kernel_size=kernel_size,
                             strides=kernel_size,
                             use_bias=False,
                             param_dtype=dtype,
                             kernel_init=nnx.with_partitioning(
                                 init_fn, (None, None, None, None, "model")),
                             rngs=rngs)

    def __call__(self, x: jax.Array) -> jax.Array:
        # x is (L, C * T * H * W)
        L, dim = x.shape
        C = dim // (self.temporal_patch_size * self.patch_size *
                    self.patch_size)
        # Reshape to (L, T, H, W, C) for Conv3D with channels_last
        x = x.reshape(L, C, self.temporal_patch_size, self.patch_size,
                      self.patch_size)
        # L,T,H,W,C
        x = jnp.transpose(x, (0, 2, 3, 4, 1))
        x = self.proj(x)
        # After conv, shape is (L, T_out, H_out, W_out, C_out)
        # With stride=kernel_size, T_out=H_out=W_out=1.
        # So shape is (L, 1, 1, 1, hidden_size)
        x = x.reshape(L, self.hidden_size)
        return x


class Qwen2_5_VisionPatchMerger(nnx.Module):

    def __init__(self, d_model: int, context_dim: int, norm_layer: Callable,
                 spatial_merge_size: int, dtype: jnp.dtype, rngs: nnx.Rngs):
        self.hidden_size = context_dim * (spatial_merge_size**2)
        self.ln_q = norm_layer(context_dim,
                               param_dtype=dtype,
                               dtype=dtype,
                               rngs=rngs,
                               scale_init=nnx.with_partitioning(
                                   init_fn, (None, )))
        self.mlp_fc1 = nnx.Linear(
            self.hidden_size,
            self.hidden_size,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, (None, "model")),
            bias_init=nnx.with_partitioning(init_fn, ("model", )),
            rngs=rngs)
        self.mlp_act = modeling_flax_utils.ACT2FN["gelu"]
        self.mlp_fc2 = nnx.Linear(
            self.hidden_size,
            d_model,
            use_bias=True,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, ("model", None)),
            bias_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rngs)

    def __call__(self, x: jax.Array) -> jax.Array:
        x = self.ln_q(x)
        x = x.reshape(-1, self.hidden_size)
        x = self.mlp_fc1(x)
        x = self.mlp_act(x)
        x = self.mlp_fc2(x)
        return x


class Qwen2_5_VisionRotaryEmbedding(nnx.Module):

    def __init__(self, dim: int, theta: float = 10000.0):
        self.dim = dim
        self.theta = theta

    def __call__(self, seqlen: int) -> jax.Array:
        inv_freq = 1.0 / (self.theta**(
            jnp.arange(0, self.dim, 2, dtype=jnp.float32) / self.dim))
        seq = jnp.arange(seqlen, dtype=jnp.float32)
        freqs = jnp.outer(seq, inv_freq)
        return freqs.astype(jnp.bfloat16)


class Qwen2_5_VisionTransformer(nnx.Module):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rngs: nnx.Rngs,
                 mesh: Mesh,
                 norm_eps: float = 1e-6):
        model_config = vllm_config.model_config
        hf_config = model_config.hf_config
        vision_config = hf_config.vision_config
        dtype = model_config.dtype

        self.config = vision_config
        self.dtype = dtype

        patch_size = vision_config.patch_size
        temporal_patch_size = vision_config.temporal_patch_size
        in_channels = vision_config.in_channels
        self.hidden_size = vision_config.hidden_size
        self.num_heads = vision_config.num_heads

        # args for get_window_index_thw
        self.window_size = vision_config.window_size
        self.patch_size = vision_config.patch_size
        self.spatial_merge_size = vision_config.spatial_merge_size
        self.fullatt_block_indexes = vision_config.fullatt_block_indexes
        self.spatial_merge_unit = self.spatial_merge_size**2

        self.patch_embed = Qwen2_5_VisionPatchEmbed(
            patch_size=patch_size,
            temporal_patch_size=temporal_patch_size,
            in_channels=in_channels,
            hidden_size=self.hidden_size,
            dtype=dtype,
            rngs=rngs)

        head_dim = vision_config.hidden_size // vision_config.num_heads
        self.rotary_pos_emb = Qwen2_5_VisionRotaryEmbedding(head_dim // 2)

        self.blocks = nnx.List([
            Qwen2_5_VisionBlock(
                config=hf_config,
                dtype=dtype,
                rngs=rngs,
                mesh=mesh,
            ) for _ in range(vision_config.depth)
        ])
        self.merger = Qwen2_5_VisionPatchMerger(
            d_model=vision_config.out_hidden_size,
            context_dim=vision_config.hidden_size,
            norm_layer=partial(nnx.RMSNorm, epsilon=norm_eps),
            spatial_merge_size=vision_config.spatial_merge_size,
            dtype=dtype,
            rngs=rngs)

        additional_config = getattr(vllm_config, "additional_config",
                                    None) or {}
        self.enable_dynamic_image_sizes = additional_config.get(
            "enable_dynamic_image_sizes", False)

    def rotary_pos_emb_thw(self, t, h, w):
        hpos_ids, wpos_ids = jnp.indices((h, w))
        hpos_ids = hpos_ids.reshape(
            h // self.spatial_merge_size,
            self.spatial_merge_size,
            w // self.spatial_merge_size,
            self.spatial_merge_size,
        ).transpose(0, 2, 1, 3).flatten()
        wpos_ids = wpos_ids.reshape(
            h // self.spatial_merge_size,
            self.spatial_merge_size,
            w // self.spatial_merge_size,
            self.spatial_merge_size,
        ).transpose(0, 2, 1, 3).flatten()
        pos_ids = jnp.stack([hpos_ids, wpos_ids], axis=-1)
        pos_ids = jnp.tile(pos_ids, (t, 1))

        max_size = max(h, w)
        rotary_pos_emb_full = self.rotary_pos_emb(max_size)

        rotary_pos_emb = rotary_pos_emb_full[pos_ids].reshape(
            pos_ids.shape[0], -1)
        rotary_pos_emb = rotary_pos_emb.reshape(
            rotary_pos_emb.shape[0] // self.spatial_merge_unit,
            self.spatial_merge_unit, -1)

        return rotary_pos_emb

    def get_window_index_thw(self, grid_t, grid_h, grid_w):
        vit_merger_window_size = (self.window_size //
                                  self.spatial_merge_size // self.patch_size)

        llm_grid_h = grid_h // self.spatial_merge_size
        llm_grid_w = grid_w // self.spatial_merge_size

        index = jnp.arange(grid_t * llm_grid_h * llm_grid_w).reshape(
            grid_t, llm_grid_h, llm_grid_w)

        pad_h = vit_merger_window_size - llm_grid_h % vit_merger_window_size
        pad_w = vit_merger_window_size - llm_grid_w % vit_merger_window_size
        num_windows_h = (llm_grid_h + pad_h) // vit_merger_window_size
        num_windows_w = (llm_grid_w + pad_w) // vit_merger_window_size

        index_padded = jnp.pad(index, ((0, 0), (0, pad_h), (0, pad_w)),
                               constant_values=-100)
        index_padded = index_padded.reshape(grid_t, num_windows_h,
                                            vit_merger_window_size,
                                            num_windows_w,
                                            vit_merger_window_size)
        index_padded = jnp.transpose(index_padded, (0, 1, 3, 2, 4)).reshape(
            grid_t, num_windows_h * num_windows_w, vit_merger_window_size,
            vit_merger_window_size)
        seqlens = (index_padded != -100).sum([2, 3]).reshape(-1)
        index_padded = index_padded.reshape(-1)
        # The number of valid indices is static because grid_t, grid_h, grid_w
        # are static.
        num_valid_indices = grid_t * llm_grid_h * llm_grid_w
        valid_indices = jnp.nonzero(index_padded != -100,
                                    size=num_valid_indices)[0]
        index_new = index_padded[valid_indices]
        cu_seqlens_tmp = jnp.cumsum(seqlens) * self.spatial_merge_unit
        cu_seqlens_tmp = cu_seqlens_tmp.astype(jnp.int32)

        # NOTE (wenlong): Pytorch code uses this to reduce replication,
        # but I don't think there is a need here, plus it would cause problem in JIT
        # Please refer here if there is a problem down-stream
        # cu_seqlens_tmp = jnp.unique(cu_seqlens_tmp)

        return index_new, cu_seqlens_tmp

    def get_rope_by_thw(self, t, h, w):
        window_index_thw, cu_seqlens_window_thw = self.get_window_index_thw(
            t, h, w)

        rotary_pos_emb_thw = self.rotary_pos_emb_thw(t, h, w)

        rotary_pos_emb_thw = rotary_pos_emb_thw[window_index_thw, :, :]
        rotary_pos_emb_thw = rotary_pos_emb_thw.reshape(
            -1, rotary_pos_emb_thw.shape[-1])
        cu_seqlens_thw = jnp.full(t, h * w, dtype=jnp.int32)

        return (rotary_pos_emb_thw, window_index_thw, cu_seqlens_window_thw,
                cu_seqlens_thw)

    def compute_attn_mask_seqlen(
        self,
        cu_seqlens: jax.Array,
    ) -> tuple[Optional[int], Optional[list[int]]]:
        max_seqlen, seqlens = None
        max_seqlen = (cu_seqlens[1:] - cu_seqlens[:-1]).max().item()
        seqlens = (cu_seqlens[1:] - cu_seqlens[:-1]).tolist()
        return max_seqlen, seqlens

    def compute_aux_arrays(self, grid_thw: tuple[tuple[int, int, int]]):
        # num of images/videoes
        num_grids = len(grid_thw)

        rotary_pos_emb = []
        window_index: list = []
        cu_window_seqlens: list = [jnp.array([0], dtype=jnp.int32)]
        cu_seqlens: list = []

        window_index_id = 0
        cu_window_seqlens_last = 0
        for i in range(num_grids):
            t, h, w = grid_thw[i]

            llm_h = h // self.spatial_merge_size
            llm_w = w // self.spatial_merge_size

            (
                rotary_pos_emb_thw,
                window_index_thw,
                cu_seqlens_window_thw,
                cu_seqlens_thw,
            ) = self.get_rope_by_thw(t, h, w)

            window_index.append(window_index_thw + window_index_id)
            window_index_id += (t * llm_h * llm_w)

            cu_seqlens_window_thw = (cu_seqlens_window_thw +
                                     cu_window_seqlens_last)
            cu_window_seqlens_last = cu_seqlens_window_thw[-1]
            cu_window_seqlens.append(cu_seqlens_window_thw)

            rotary_pos_emb.append(rotary_pos_emb_thw)

            cu_seqlens.append(cu_seqlens_thw)

        rotary_pos_emb = jnp.concatenate(rotary_pos_emb, axis=0)
        window_index = jnp.concatenate(window_index, axis=0)
        cu_window_seqlens = jnp.concatenate(cu_window_seqlens, axis=0)

        cu_seqlens = jnp.concatenate(cu_seqlens, axis=0)
        cu_seqlens = jnp.cumsum(cu_seqlens, axis=0, dtype=jnp.int32)
        cu_seqlens = jnp.pad(cu_seqlens, ((1, 0), ),
                             mode='constant',
                             constant_values=0)
        return window_index, rotary_pos_emb, cu_seqlens, cu_window_seqlens

    def pad_inputs(self, x, window_index, rotary_pos_emb, cu_seqlens,
                   cu_window_seqlens):
        # padding
        num_patches = int(rotary_pos_emb.shape[0])
        bucket_num_patches = 1 << (num_patches - 1).bit_length()
        num_tokens = window_index.shape[0]
        bucket_num_tokens = bucket_num_patches // self.spatial_merge_unit
        vit_merger_window_size = (self.window_size //
                                  self.spatial_merge_size // self.patch_size)
        max_windows = (bucket_num_tokens // vit_merger_window_size) + 2

        rotary_pos_emb = jnp.pad(rotary_pos_emb,
                                 ((0, bucket_num_patches - num_patches),
                                  (0, 0)))
        window_index = jnp.concatenate([
            window_index,
            jnp.arange(num_tokens, bucket_num_tokens, dtype=jnp.int32)
        ])
        cu_window_seqlens = jnp.append(cu_window_seqlens, bucket_num_patches)
        pad_w = max(0, max_windows + 1 - cu_window_seqlens.shape[0])
        cu_window_seqlens = jnp.pad(cu_window_seqlens, (0, pad_w), mode='edge')
        cu_seqlens = jnp.append(cu_seqlens, bucket_num_patches)

        x_padded = jnp.pad(x, ((0, bucket_num_patches - x.shape[0]), (0, 0)))

        return x_padded, window_index, rotary_pos_emb, cu_seqlens, cu_window_seqlens, num_tokens

    def compute_hidden_states(self, x: jax.Array, window_index: jax.Array,
                              rotary_pos_emb: jax.Array, cu_seqlens: jax.Array,
                              cu_window_seqlens: jax.Array) -> jax.Array:
        hidden_states = self.patch_embed(x)

        # num of patches
        seq_len = x.shape[0]

        hidden_states = hidden_states.reshape(
            seq_len // self.spatial_merge_unit, self.spatial_merge_unit, -1)
        hidden_states = hidden_states[window_index, :, :]
        hidden_states = hidden_states.reshape(seq_len, -1)

        hidden_states = jnp.expand_dims(hidden_states, axis=1)

        for layer_num, blk in enumerate(self.blocks):
            if layer_num in self.fullatt_block_indexes:
                hidden_states = blk(hidden_states,
                                    rotary_pos_emb=rotary_pos_emb,
                                    cu_window_seqlens=cu_seqlens,
                                    use_fullattn=True)
            else:
                hidden_states = blk(hidden_states,
                                    rotary_pos_emb=rotary_pos_emb,
                                    cu_window_seqlens=cu_window_seqlens,
                                    use_fullattn=False)

        # adapter
        hidden_states = self.merger(hidden_states)
        reverse_indices = jnp.argsort(window_index)
        hidden_states = hidden_states[reverse_indices, :]
        return hidden_states

    @jax.jit
    def encode_padded_jit(self, x_padded, window_index, rotary_pos_emb,
                          cu_seqlens, cu_window_seqlens):
        return self.compute_hidden_states(x_padded, window_index,
                                          rotary_pos_emb, cu_seqlens,
                                          cu_window_seqlens)

    @jax.jit(static_argnames=("grid_thw", ))
    def encode_jit(self, x, grid_thw):
        window_index, rotary_pos_emb, cu_seqlens, cu_window_seqlens = self.compute_aux_arrays(
            grid_thw)
        return self.compute_hidden_states(x, window_index, rotary_pos_emb,
                                          cu_seqlens, cu_window_seqlens)

    def __call__(self, x: jax.Array, grid_thw: tuple[tuple[int, int,
                                                           int]]) -> jax.Array:
        # x: pixel_values: jax.Array
        # """Shape:
        # `(num_patches, num_channels * patch_size * patch_size)`
        # """

        # grid_thw: image_grid_thw: jax.Array
        # """Shape: `(num_images, 3)`
        # This should be in `(grid_t, grid_h, grid_w)` format.
        # """
        if self.enable_dynamic_image_sizes:
            window_index, rotary_pos_emb, cu_seqlens, cu_window_seqlens = self.compute_aux_arrays(
                grid_thw)
            x_padded, window_index, rotary_pos_emb, cu_seqlens, cu_window_seqlens, num_tokens = self.pad_inputs(
                x, window_index, rotary_pos_emb, cu_seqlens, cu_window_seqlens)

            hidden_states = self.encode_padded_jit(x_padded, window_index,
                                                   rotary_pos_emb, cu_seqlens,
                                                   cu_window_seqlens)
            return hidden_states[:num_tokens]

        else:
            return self.encode_jit(x, grid_thw)


class Qwen2_5_VLForConditionalGeneration(nnx.Module):
    WeightLoader = StandardWeightLoader

    def __init__(self, vllm_config: VllmConfig, rng_key: jax.Array,
                 mesh: Mesh) -> None:
        config: Qwen2_5_VLConfig = vllm_config.model_config.hf_config
        multimodal_config = vllm_config.model_config.multimodal_config

        self.vllm_config = vllm_config
        self.rng = nnx.Rngs(rng_key)
        self.mesh = mesh
        self.is_first_rank = get_pp_group().is_first_rank

        self.config = config
        self.multimodal_config = multimodal_config

        self.is_last_rank = get_pp_group().is_last_rank

        if self.is_first_rank:
            self.visual = Qwen2_5_VisionTransformer(
                vllm_config=vllm_config,
                rngs=self.rng,
                mesh=mesh,
                norm_eps=getattr(config, "rms_norm_eps", 1e-6),
            )
        else:
            self.visual = PPMissingLayer()

        self.model = Qwen2Model(vllm_config, self.rng, mesh)
        model_config = vllm_config.model_config
        if not model_config.hf_config.tie_word_embeddings:
            if self.is_last_rank:
                vocab_size = model_config.get_vocab_size()
                hidden_size = model_config.hf_config.hidden_size
                self.lm_head = JaxEinsum(
                    einsum_str="TD,DV->TV",
                    kernel_shape=(hidden_size, vocab_size),
                    dtype=model_config.dtype,
                    rngs=self.rng,
                    quant_config=vllm_config.quant_config,
                )
            else:
                self.lm_head = PPMissingLayer()

    def get_mrope_input_positions(
        self,
        input_tokens: list[int],
        hf_config,
        image_grid_thw,
        video_grid_thw,
        second_per_grid_ts: list[float],
        context_len: int = 0,
        seq_len: int | None = None,
        audio_feature_lengths=None,
        use_audio_in_video: bool = False,
    ) -> tuple[jax.Array, int]:
        """Get mrope input positions and delta value."""

        image_token_id = hf_config.image_token_id
        video_token_id = hf_config.video_token_id
        vision_start_token_id = hf_config.vision_start_token_id
        spatial_merge_size = hf_config.vision_config.spatial_merge_size
        tokens_per_second = getattr(hf_config.vision_config,
                                    "tokens_per_second", 1.0)

        input_tokens_tensor = np.array(input_tokens)
        vision_start_indices = np.argwhere(
            input_tokens_tensor == vision_start_token_id).squeeze(1)
        vision_tokens = input_tokens_tensor[vision_start_indices + 1]
        image_nums = np.sum(vision_tokens == image_token_id)
        video_nums = np.sum(vision_tokens == video_token_id)
        llm_pos_ids_list: list = []

        st = 0
        remain_images, remain_videos = image_nums, video_nums

        image_index, video_index = 0, 0
        for _ in range(image_nums + video_nums):
            video_second_per_grid_t = 0.0
            if remain_images > 0:
                try:
                    ed_image = input_tokens.index(image_token_id, st)
                except ValueError:
                    ed_image = len(input_tokens) + 1
            else:
                ed_image = len(input_tokens) + 1
            if remain_videos > 0:
                try:
                    ed_video = input_tokens.index(video_token_id, st)
                except ValueError:
                    ed_video = len(input_tokens) + 1
            else:
                ed_video = len(input_tokens) + 1
            if ed_image < ed_video:
                t, h, w = (
                    image_grid_thw[image_index][0],
                    image_grid_thw[image_index][1],
                    image_grid_thw[image_index][2],
                )
                image_index += 1
                remain_images -= 1
                ed = ed_image
            else:
                t, h, w = (
                    video_grid_thw[video_index][0],
                    video_grid_thw[video_index][1],
                    video_grid_thw[video_index][2],
                )
                video_second_per_grid_t = 1.0
                if second_per_grid_ts:
                    video_second_per_grid_t = second_per_grid_ts[video_index]
                video_index += 1
                remain_videos -= 1
                ed = ed_video

            llm_grid_t, llm_grid_h, llm_grid_w = (
                t,
                h // spatial_merge_size,
                w // spatial_merge_size,
            )
            text_len = ed - st

            st_idx = llm_pos_ids_list[-1].max().item() + 1 if len(
                llm_pos_ids_list) > 0 else 0
            llm_pos_ids_list.append(
                jnp.broadcast_to(
                    jnp.arange(text_len, dtype=jnp.int32).reshape(1, -1),
                    (3, text_len)) + st_idx)

            t_index = ((jnp.broadcast_to(
                jnp.arange(llm_grid_t, dtype=jnp.int32).reshape(-1, 1),
                (llm_grid_t, llm_grid_h * llm_grid_w)) *
                        video_second_per_grid_t * tokens_per_second).astype(
                            jnp.int32).flatten())

            h_index = (jnp.broadcast_to(
                jnp.arange(llm_grid_h, dtype=jnp.int32).reshape(1, -1, 1),
                (llm_grid_t, llm_grid_h, llm_grid_w)).flatten())
            w_index = (jnp.broadcast_to(
                jnp.arange(llm_grid_w, dtype=jnp.int32).reshape(1, 1, -1),
                (llm_grid_t, llm_grid_h, llm_grid_w)).flatten())

            llm_pos_ids_list.append(
                jnp.stack([t_index, h_index, w_index]) + text_len + st_idx)
            st = ed + llm_grid_t * llm_grid_h * llm_grid_w

        if st < len(input_tokens):
            st_idx = llm_pos_ids_list[-1].max().item() + 1 if len(
                llm_pos_ids_list) > 0 else 0
            text_len = len(input_tokens) - st

            llm_pos_ids_list.append(
                jnp.broadcast_to(
                    jnp.arange(text_len, dtype=jnp.int32).reshape(1, -1),
                    (3, text_len)) + st_idx)

        llm_positions = jnp.concatenate(llm_pos_ids_list,
                                        axis=1).reshape(3, -1)
        mrope_position_delta = (llm_positions.max() + 1 -
                                len(input_tokens)).item()
        llm_positions = llm_positions[:, context_len:seq_len]

        return llm_positions, mrope_position_delta

    def _validate_and_reshape_mm_tensor(self, mm_input: object,
                                        name: str) -> jax.Array:
        if isinstance(mm_input, list):
            # Assuming it's a list of arrays (e.g., np.ndarray, torch.Tensor)
            # that can be concatenated.
            arrays_to_concat = [jnp.asarray(item) for item in mm_input]
            return jnp.concatenate(arrays_to_concat, axis=0)

        # Handle single array-like objects (np.ndarray, torch.Tensor, jax.Array)
        if hasattr(mm_input, 'ndim'):
            array_input = jnp.asarray(mm_input)
            if array_input.ndim == 2:
                return array_input
            if array_input.ndim == 3:
                # This reshapes the batched 3D tensor to a 2D tensor.
                return array_input.reshape(-1, array_input.shape[-1])

        raise ValueError(f"Incorrect type of {name}. "
                         f"Got type: {type(mm_input)}")

    def _parse_and_validate_image_input(
            self, image_grid_thw: tuple[tuple[int, int, int], ...],
            **kwargs: object) -> Optional[Qwen2_5_VLImageInputs]:
        pixel_values = kwargs.pop("pixel_values", None)
        image_embeds = kwargs.pop("image_embeds", None)
        # image_grid_thw = kwargs.pop("image_grid_thw", None)

        if pixel_values is None and image_embeds is None:
            return None

        if pixel_values is not None:
            pixel_values = self._validate_and_reshape_mm_tensor(
                pixel_values, "image pixel values")
            # image_grid_thw = self._validate_and_reshape_mm_tensor(
            #     image_grid_thw, "image grid_thw")

            if not isinstance(pixel_values, jax.Array):
                raise ValueError("Incorrect type of image pixel values. "
                                 f"Got type: {type(pixel_values)}")

            return Qwen2_5_VLImagePixelInputs(type="pixel_values",
                                              pixel_values=pixel_values,
                                              image_grid_thw=image_grid_thw)

        # Note: comment them out for now and save for future support
        # if image_embeds is not None:
        #     image_embeds = self._validate_and_reshape_mm_tensor(
        #         image_embeds, "image embeds")
        #     image_grid_thw = self._validate_and_reshape_mm_tensor(
        #         image_grid_thw, "image grid_thw")

        #     if not isinstance(image_embeds, jax.Array):
        #         raise ValueError("Incorrect type of image embeddings. "
        #                          f"Got type: {type(image_embeds)}")
        #     return Qwen2_5_VLImageEmbeddingInputs(
        #         type="image_embeds",
        #         image_embeds=image_embeds,
        #         image_grid_thw=image_grid_thw)

    def _parse_and_validate_multimodal_inputs(self,
                                              image_grid_thw: tuple[tuple[int,
                                                                          int,
                                                                          int],
                                                                    ...],
                                              **kwargs: object) -> dict:
        mm_input_by_modality = {}

        # Preserve the order of modalities if there are multiple of them
        # from the order of kwargs.
        for input_key in kwargs:
            if input_key in ("pixel_values", "image_embeds"
                             ) and "image" not in mm_input_by_modality:
                mm_input_by_modality[
                    "image"] = self._parse_and_validate_image_input(
                        image_grid_thw, **kwargs)
            # if input_key in ("pixel_values_videos", "video_embeds"
            #                  ) and "video" not in mm_input_by_modality:
            #     mm_input_by_modality[
            #         "video"] = self._parse_and_validate_video_input(**kwargs)
        return mm_input_by_modality

    def get_single_image_embedding(self, image_pixel_values, image_grid_thw):
        return self.visual(image_pixel_values, (image_grid_thw, ))

    def _process_image_input(
            self, image_input: Qwen2_5_VLImageInputs) -> tuple[jax.Array, ...]:

        grid_thw = image_input["image_grid_thw"]

        if image_input["type"] == "image_embeds":
            image_embeds = image_input["image_embeds"].astype(
                self.visual.dtype)
        else:
            pixel_values = image_input["pixel_values"]
            image_embeds = []
            current_idx = 0
            for image_thw in grid_thw:
                t, h, w = image_thw
                image_size = t * h * w
                end_idx = current_idx + image_size
                image_pixel_values = pixel_values[current_idx:end_idx, :]
                image_embeds.append(
                    self.get_single_image_embedding(image_pixel_values,
                                                    image_thw))
                current_idx = end_idx
            image_embeds = jnp.concatenate(image_embeds, axis=0)

        # Split concatenated embeddings for each image item.
        merge_size = self.visual.config.spatial_merge_size
        sizes = np.prod(np.array(grid_thw, dtype=np.int64),
                        axis=-1) // merge_size // merge_size

        if sizes.size == 0:
            return ()
        if sizes.size == 1:
            return (image_embeds, )

        split_indices = np.cumsum(sizes)[:-1]
        return tuple(jnp.split(image_embeds, split_indices))

    def embed_multimodal(self, image_grid_thw: tuple[tuple[int, int, int],
                                                     ...],
                         **kwargs: object) -> MultiModalEmbeddings:

        if not self.is_first_rank:
            return ()

        mm_input_by_modality = self._parse_and_validate_multimodal_inputs(
            image_grid_thw, **kwargs)
        if not mm_input_by_modality:
            return []

        # The result multimodal_embeddings is tuple of tensors, with each
        # tensor correspoending to a multimodal data item (image or video).
        multimodal_embeddings: tuple[jax.Array, ...] = ()

        # NOTE: It is important to iterate over the keys in this dictionary
        # to preserve the order of the modalities.
        for modality in mm_input_by_modality:
            multimodal_input = mm_input_by_modality[modality]
            if modality == "image":
                vision_embeddings = self._process_image_input(multimodal_input)
                multimodal_embeddings += vision_embeddings
            # if modality == "video":
            #     video_embeddings = self._process_video_input(multimodal_input)
            #     multimodal_embeddings += video_embeddings

        return multimodal_embeddings

    def embed_input_ids(
        self,
        input_ids: jax.Array,
        multimodal_embeddings: jax.Array | None,
        *,
        is_multimodal: jax.Array | None = None,
    ) -> jax.Array:

        del is_multimodal

        if not self.is_first_rank:
            return None

        inputs_embeds = self.model.embed_tokens(input_ids)


        if multimodal_embeddings is not None \
            and multimodal_embeddings.shape[0] != 0:
            inputs_embeds = merge_multimodal_embeddings(
                input_ids, inputs_embeds, multimodal_embeddings,
                [self.config.image_token_id, self.config.video_token_id])

        return inputs_embeds

    def __call__(
        self,
        kv_caches: list[jax.Array],
        input_ids: Optional[jax.Array],
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
        _input_positions=None,
        _layer_name_to_kv_cache=None,
        _lora_metadata=None,
        intermediate_tensors: Optional[JaxIntermediateTensors] = None,
        is_first_rank: bool = True,
        is_last_rank: bool = True,
        *args,
    ) -> tuple[list[jax.Array], jax.Array | JaxIntermediateTensors,
               List[jax.Array]]:
        # The logic of choosing between input_ids and inputs_embeds is
        # handled inside self.model.__call__
        if not is_first_rank:
            assert intermediate_tensors is not None
            inputs_embeds = intermediate_tensors["hidden_states"]

        kv_caches, x = self.model(
            kv_caches=kv_caches,
            input_ids=input_ids,
            attention_metadata=attention_metadata,
            inputs_embeds=inputs_embeds,
        )

        if not is_last_rank:
            x = JaxIntermediateTensors(tensors={"hidden_states": x})

        return kv_caches, x, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        if hasattr(self, 'lm_head'):
            return self.lm_head(hidden_states)

        return self.model.embed_tokens.decode(hidden_states)

    def load_weights(self, rng_key: jax.Array) -> None:
        self.rng = nnx.Rngs(rng_key)
        # Key: path to a HF layer weight
        # Value: a tuple of (path to a nnx layer weight, nnx weight sharding)

        self.pp_missing_layers = []
        for path, module in nnx.iter_graph(self):
            if isinstance(module, PPMissingLayer):
                layer_name = ".".join([str(s) for s in path])
                self.pp_missing_layers.append(layer_name)

        mappings = {
            "visual.blocks.*.attn.proj.bias": "visual.blocks.*.attn.proj.bias",
            "visual.blocks.*.attn.proj": "visual.blocks.*.attn.proj.kernel",
            "visual.blocks.*.attn.qkv.bias":
            "visual.blocks.*.attn.qkv_proj.bias",
            "visual.blocks.*.attn.qkv": "visual.blocks.*.attn.qkv_proj.kernel",
            "visual.blocks.*.mlp.down_proj.bias":
            "visual.blocks.*.mlp.down_proj.bias",
            "visual.blocks.*.mlp.down_proj":
            "visual.blocks.*.mlp.down_proj.kernel",
            "visual.blocks.*.mlp.gate_proj.bias":
            "visual.blocks.*.mlp.gate_proj.bias",
            "visual.blocks.*.mlp.gate_proj":
            "visual.blocks.*.mlp.gate_proj.kernel",
            "visual.blocks.*.mlp.up_proj.bias":
            "visual.blocks.*.mlp.up_proj.bias",
            "visual.blocks.*.mlp.up_proj":
            "visual.blocks.*.mlp.up_proj.kernel",
            "visual.blocks.*.norm1": "visual.blocks.*.norm1.scale",
            "visual.blocks.*.norm2": "visual.blocks.*.norm2.scale",
            "visual.merger.ln_q": "visual.merger.ln_q.scale",
            "visual.merger.mlp.0.bias": "visual.merger.mlp_fc1.bias",
            "visual.merger.mlp.0": "visual.merger.mlp_fc1.kernel",
            "visual.merger.mlp.2.bias": "visual.merger.mlp_fc2.bias",
            "visual.merger.mlp.2": "visual.merger.mlp_fc2.kernel",
            "visual.patch_embed.proj": "visual.patch_embed.proj.kernel",
        }

        loader = self.WeightLoader(self.vllm_config, self.mesh)
        keep_hf_weight_suffix_when_match = ['model']
        if not self.vllm_config.model_config.hf_config.tie_word_embeddings:
            keep_hf_weight_suffix_when_match.append('lm_head')
        loader.load_weights(
            self,
            mappings,
            keep_hf_weight_suffix_when_match=keep_hf_weight_suffix_when_match)

    def precompile_vision_encoder(
        self,
        run_compilation_fn: Callable,
    ) -> None:
        if not self.is_first_rank:
            return

        vc = self.vllm_config.model_config.hf_config.vision_config
        patch_input_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        if self.visual.enable_dynamic_image_sizes:
            spatial_merge_unit = vc.spatial_merge_size**2
            max_num_batched_tokens = self.vllm_config.scheduler_config.max_num_batched_tokens
            mm_kwargs = self.vllm_config.model_config.multimodal_config.mm_processor_kwargs or {}
            # Use size.longest_edge if provided, otherwise default to inf
            if "size" in mm_kwargs and "longest_edge" in mm_kwargs.get(
                    "size", {}):
                limit_pixels = float(mm_kwargs["size"]["longest_edge"])
            else:
                limit_pixels = float('inf')

            max_patches = int(
                min(max_num_batched_tokens * spatial_merge_unit,
                    limit_pixels / (vc.patch_size**2)))

            num_patches_paddings = [
                1 << i for i in range(4, (max_patches - 1).bit_length() + 1)
            ]
            rotary_dim = vc.hidden_size // vc.num_heads // 2
            vit_merger_window_size = (vc.window_size //
                                      vc.spatial_merge_size // vc.patch_size)

            for num_patches in num_patches_paddings:
                dummy_x_padded = jnp.ones(
                    (num_patches, patch_input_dim),
                    dtype=self.vllm_config.model_config.dtype)

                num_tokens = num_patches // spatial_merge_unit
                dummy_window_index = jnp.arange(num_tokens, dtype=jnp.int32)

                dummy_rotary_pos_emb = jnp.ones(
                    (num_patches, rotary_dim),
                    dtype=self.vllm_config.model_config.dtype)

                dummy_cu_seqlens = jnp.array([0, num_patches, num_patches],
                                             dtype=jnp.int32)

                max_windows = (num_tokens // vit_merger_window_size) + 2
                patches_per_window = (vit_merger_window_size**
                                      2) * spatial_merge_unit
                dummy_cu_window_seqlens = jnp.arange(
                    max_windows + 1, dtype=jnp.int32) * patches_per_window
                dummy_cu_window_seqlens = jnp.minimum(dummy_cu_window_seqlens,
                                                      num_patches)

                run_compilation_fn("vision_encoder_padded",
                                   self.visual.encode_padded_jit,
                                   dummy_x_padded,
                                   dummy_window_index,
                                   dummy_rotary_pos_emb,
                                   dummy_cu_seqlens,
                                   dummy_cu_window_seqlens,
                                   num_patches=num_patches)
        else:
            image_shapes = []
            if (warmup_config := self.vllm_config.additional_config.get(
                    "vision_warmup_config")):
                image_shapes = warmup_config.get("image_shapes")

            factor = vc.patch_size * vc.spatial_merge_size
            for input_hw in image_shapes:
                if not isinstance(input_hw, list) or len(input_hw) != 2:
                    logger.warning(f"Skipping invalid shape {input_hw}.")
                    continue
                h_input, w_input = input_hw
                h_processed = round(h_input / factor) * factor
                w_processed = round(w_input / factor) * factor
                t, h, w = 1, h_processed // vc.patch_size, w_processed // vc.patch_size
                grid_thw = (t, h, w)
                num_patches = t * h * w

                dummy_pixel_values = jnp.ones(
                    (num_patches, patch_input_dim),
                    self.vllm_config.model_config.dtype,
                )
                dummy_grid_thw = (grid_thw, )

                run_compilation_fn("vision_encoder",
                                   self.visual.encode_jit,
                                   dummy_pixel_values,
                                   dummy_grid_thw,
                                   image_shape=input_hw)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
import torchax
from jax.sharding import Mesh
from vllm.v1.attention.backend import AttentionType

from tpu_inference.layers.vllm.custom_ops.mla_attention import (
    VllmMLAAttention, VllmMultiHeadLatentAttentionWrapper)
from tpu_inference.models.vllm.vllm_model_wrapper_context import (
    get_vllm_model_wrapper_context, set_vllm_model_wrapper_context)


@pytest.fixture
def mesh():
    """Provides a mock 1D JAX mesh for testing."""
    devices = np.array(jax.local_devices())[0:1]
    if not devices.any():
        devices = np.array([jax.devices("cpu")[0]])
    return Mesh(devices.reshape((-1, 1, 1)), ("data", "attn_dp", "model"))


class TestVllmTPUMLAAttention:

    @patch(
        "vllm.model_executor.layers.attention.mla_attention.MLAAttention.__init__",
        autospec=True)
    def test_init_auto_dtype(self, mock_super_init):
        # Emulate how vLLM initializes the cache type
        def side_effect(self, *args, **kwargs):
            self.kv_cache_dtype = "auto"

        mock_super_init.side_effect = side_effect

        kv_b_proj = MagicMock()
        attn = VllmMLAAttention(
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            kv_b_proj=kv_b_proj,
            prefix="test",
        )

        # Validate that the fallback attributes are set correctly
        assert attn.kv_sharing_target_layer_name is None
        assert attn.attn_type == AttentionType.DECODER
        assert attn.sliding_window is None
        assert attn.kv_cache_quantized_dtype is None

    @patch(
        "vllm.model_executor.layers.attention.mla_attention.MLAAttention.__init__",
        autospec=True)
    def test_init_fp8_dtype(self, mock_super_init):

        def side_effect(self, *args, **kwargs):
            self.kv_cache_dtype = "fp8"

        mock_super_init.side_effect = side_effect

        kv_b_proj = MagicMock()
        attn = VllmMLAAttention(
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            kv_b_proj=kv_b_proj,
            prefix="test",
        )

        assert attn.kv_cache_quantized_dtype == jnp.float8_e4m3fn

    @patch(
        "vllm.model_executor.layers.attention.mla_attention.MLAAttention.__init__",
        autospec=True)
    def test_process_weights_after_loading(self, mock_super_init, mesh):

        def side_effect(self, *args, **kwargs):
            self.kv_cache_dtype = "auto"

        mock_super_init.side_effect = side_effect

        # Mock a linear layer (to simulate column parallel linear)
        kv_b_proj = torch.nn.Linear(10, 10)
        attn = VllmMLAAttention(
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            kv_b_proj=kv_b_proj,
            prefix="test",
        )

        attn.W_UK_T = torchax.tensor.Tensor(jnp.ones((10, 10)),
                                            env=torchax.default_env())
        attn.W_UV = torchax.tensor.Tensor(jnp.ones((10, 10)),
                                          env=torchax.default_env())
        attn.kv_b_proj = MagicMock()
        attn.kv_b_proj.quant_method.linear_config.mesh = mesh
        attn.kv_b_proj.named_parameters.return_value = {}

        with patch(
                "vllm.model_executor.layers.attention.mla_attention.MLAAttention.process_weights_after_loading"
        ) as mock_super_process:
            with torchax.default_env():
                attn.process_weights_after_loading(torch.float32)

            mock_super_process.assert_called_once_with(torch.float32)

            # W_UK_T and W_UV should now be un-grad Parameters
            assert isinstance(attn.W_UK_T, torch.nn.Parameter)
            assert not attn.W_UK_T.requires_grad
            assert isinstance(attn.W_UV, torch.nn.Parameter)
            assert not attn.W_UV.requires_grad

    @patch(
        "tpu_inference.layers.vllm.custom_ops.mla_attention.get_attention_context"
    )
    @patch(
        "vllm.model_executor.layers.attention.mla_attention.MLAAttention.__init__",
        autospec=True)
    def test_forward(self, mock_super_init, mock_get_attention_context, mesh):

        def side_effect(self, *args, **kwargs):
            self.kv_cache_dtype = "auto"

        mock_super_init.side_effect = side_effect

        attn = VllmMLAAttention(
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            kv_b_proj=MagicMock(),
            prefix="test_layer",
        )

        attn.calculate_kv_scales = False
        attn.layer_name = "test_layer"
        attn.impl = MagicMock()

        mock_outputs = jnp.zeros((1, 10))
        mock_new_kv_cache = jnp.zeros((10, 10))
        attn.impl.forward.return_value = (mock_outputs, mock_new_kv_cache)

        mock_attn_metadata = MagicMock()
        mock_get_attention_context.return_value = (mock_attn_metadata, None,
                                                   None, None)

        q = torch.rand(1, 10)
        kv_c_normed = torch.rand(1, 10)
        k_pe = torch.rand(1, 10)

        kv_cache = jnp.zeros((10, 10))

        with set_vllm_model_wrapper_context(
                kv_caches=[kv_cache],
                mesh=mesh,
                layer_name_to_kvcache_index={"test_layer": 0}):
            outputs = attn.forward(q, kv_c_normed, k_pe)

            mock_get_attention_context.assert_called_once_with("test_layer")
            attn.impl.forward.assert_called_once_with(q, kv_c_normed, k_pe,
                                                      kv_cache,
                                                      mock_attn_metadata, mesh,
                                                      attn)

            assert isinstance(outputs, torch.Tensor)
            context = get_vllm_model_wrapper_context()
            assert context.kv_caches[0] is mock_new_kv_cache

    @patch("torch.ops.vllm.maybe_calc_kv_scales", create=True)
    @patch(
        "tpu_inference.layers.vllm.custom_ops.mla_attention.get_attention_context"
    )
    @patch(
        "vllm.model_executor.layers.attention.mla_attention.MLAAttention.__init__",
        autospec=True)
    def test_forward_calculates_kv_scales(self, mock_super_init,
                                          mock_get_attention_context,
                                          mock_maybe_calc_kv_scales, mesh):

        def side_effect(self, *args, **kwargs):
            self.kv_cache_dtype = "auto"

        mock_super_init.side_effect = side_effect

        attn = VllmMLAAttention(
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            kv_b_proj=MagicMock(),
            prefix="test_layer",
        )

        attn.calculate_kv_scales = True
        attn.layer_name = "test_layer"
        attn.impl = MagicMock()
        attn.impl.forward.return_value = (jnp.zeros(1), jnp.zeros(1))
        mock_get_attention_context.return_value = (MagicMock(), None, None,
                                                   None)

        q = torch.rand(1, 10)
        kv_c_normed = torch.rand(1, 10)
        k_pe = torch.rand(1, 10)

        with set_vllm_model_wrapper_context(
                kv_caches=[jnp.zeros(1)],
                mesh=mesh,
                layer_name_to_kvcache_index={"test_layer": 0}):
            attn.forward(q, kv_c_normed, k_pe)

            mock_maybe_calc_kv_scales.assert_called_once_with(
                q, kv_c_normed, k_pe, "test_layer")


class TestVllmTPUMultiHeadLatentAttentionWrapper:

    @patch(
        "tpu_inference.layers.vllm.custom_ops.mla_attention.VllmMLAAttention",
        autospec=True)
    def test_init(self, mock_tpu_mla_attn):
        mla_modules = MagicMock()
        mla_modules.fused_qkv_a_proj = "fused_qkv_a_proj"
        mla_modules.kv_a_proj_with_mqa = "kv_a_proj_with_mqa"
        mla_modules.q_a_layernorm = "q_a_layernorm"
        mla_modules.q_b_proj = "q_b_proj"
        mla_modules.q_proj = "q_proj"
        mla_modules.kv_a_layernorm = "kv_a_layernorm"
        mla_modules.kv_b_proj = MagicMock()
        mla_modules.rotary_emb = "rotary_emb"
        mla_modules.o_proj = "o_proj"
        mla_modules.indexer = None
        mla_modules.indexer_rotary_emb = "indexer_rotary_emb"
        mla_modules.is_sparse = False

        wrapper = VllmMultiHeadLatentAttentionWrapper(
            hidden_size=128,
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            mla_modules=mla_modules,
            cache_config=None,
            quant_config=None,
            prefix="test_wrapper",
        )

        assert wrapper.hidden_size == 128
        assert wrapper.qk_head_dim == 96
        assert wrapper.fused_qkv_a_proj == "fused_qkv_a_proj"
        assert wrapper.kv_a_proj_with_mqa == "kv_a_proj_with_mqa"
        assert wrapper.q_a_layernorm == "q_a_layernorm"
        assert wrapper.q_b_proj == "q_b_proj"
        assert wrapper.q_proj == "q_proj"
        assert wrapper.kv_a_layernorm == "kv_a_layernorm"
        assert wrapper.kv_b_proj == mla_modules.kv_b_proj
        assert wrapper.rotary_emb == "rotary_emb"
        assert wrapper.o_proj == "o_proj"
        assert wrapper.indexer is None
        assert wrapper.is_sparse is False

        mock_tpu_mla_attn.assert_called_once_with(
            num_heads=8,
            scale=1.0,
            qk_nope_head_dim=64,
            qk_rope_head_dim=32,
            v_head_dim=64,
            q_lora_rank=None,
            kv_lora_rank=32,
            cache_config=None,
            quant_config=None,
            prefix="test_wrapper.attn",
            kv_b_proj=mla_modules.kv_b_proj,
            use_sparse=False,
            indexer=None,
        )
        assert wrapper.mla_attn == mock_tpu_mla_attn.return_value
